import numpy as np
from numpy.random import Generator
from typing import Type, Any
from .target import Target
from .scale import Scale 
from ..utils.general_utils import validate_and_convert_location
from ..utils import montecarlo as mc
from ..utils.custom_types import FloatLike, PairOfFloats
from .morphology import Morphology
from abc import ABC, abstractmethod

class Impact(ABC):
    """
    Abstract base class representing an impact event in the simulation. 

    This class provides a generic framework for impact-related entities, defining common attributes and methods such as diameter,
    radius, and location. It is not intended to be instantiated directly but serves as a base for specific impact event classes like 
    Crater and Projectile.


    Parameters
    ----------
    diameter : FloatLike, optional
        The diameter of the impact event in meters.
    radius : FloatLike, optional
        The radius of the impact event in meters.
    location : PairOfFloats, optional
        The location (latitude, longitude) of the impact event.
    age : FloatLike, optional
        The age of the impact event in My before present.
    target : Target, optional
        The target body of the impact event.
    scale_cls : Type[Scale], optional
        The class used for scaling calculations.
    rng : Generator, optional
        Random number generator instance.

    Abstract Methods
    ----------------
    diameter
        Property that must be implemented to handle the diameter of the impact event.
    radius
        Property that must be implemented to handle the radius of the impact event.

    Notes
    -----
    Subclasses must provide implementations for all abstract methods and properties. 
    They may also extend the __init__ method to handle additional parameters specific 
    to the subclass.

    Examples
    --------
    Subclassing `Impact` to create a `Crater` class:

    >>> class Crater(Impact):
    ...     def __init__(self, diameter, ...):
    ...         super().__init__(diameter=diameter, ...)
    ...     # Implement abstract methods and properties here

    See Also
    --------
    Crater, Projectile : Subclasses of `Impact`.
    """
    
    def __init__(self, 
                 diameter: FloatLike = None,
                 radius: FloatLike = None,
                 location: PairOfFloats = None,
                 age: FloatLike = None,
                 target: Target = None,
                 scale_cls: Type[Scale] = None,
                 rng: Generator = None,
                 **kwargs: Any):
        
        # Evaluate and check diameter/radius inputs
        self._target = None
        self._rng = None
        self._scale_cls = None
        self._scale = None
        self._location = None
        self._age = None
        self._diameter = None
        self._radius = None
        self._face_index = None
        self._node_index = None
        
        self.target = target
        self.rng = rng 
        self.scale_cls = scale_cls 
        self.scale = self.scale_cls(target=self.target, rng=self.rng,**kwargs)
        self.location = location 
        self.age = age
        self.diameter = diameter
        self.radius = radius
        
        values_set = sum(x is not None for x in [diameter, radius])
        if values_set > 1:
            raise ValueError("Only one of diameter or radius may be set") 

    @property
    @abstractmethod
    def diameter(self):
        """
        The diameter of the the impact object in m.
        
        Returns
        -------
        np.float64
        """         
        pass

    @property
    @abstractmethod
    def radius(self):
        """
        The radius of the the impact object in m.
        
        Returns
        -------
        np.float64
        """             
        pass

    @property
    def location(self):
        """
        The longitude and latitude of the impact in degrees. 
       
        Returns
        -------
        np.ndarray
        """
        return self._location

    @location.setter
    def location(self, value):
        if value is None:
            self._location = mc.get_random_location(rng=self.rng)
        else:   
            if len(value) == 1:
                value = value.item() 
            self._location = validate_and_convert_location(value)
           
    @property
    def age(self):
        """
        The age of the impact in My before present.
        
        Returns
        -------
        np.float64
        """
        return self._age
    
    @age.setter
    def age(self, value):
        if value is None:
            self._age = None
        else:
            self._age = np.float64(value)
        return
            
    @property
    def face_index(self):
        """
        The index of the face closest to the impact location.
        
        Returns
        -------
        int
        """
        return self._face_index
    
    @face_index.setter
    def face_index(self, value):
        if value is None:
            self._face_index = None
        else:
            self._face_index = int(value)
        return
    
    @property
    def node_index(self):
        """
        The index of the node closest to the impact location.
        
        Returns
        -------
        int
        """
        return self._node_index
    
    @node_index.setter
    def node_index(self, value):
        if value is None:
            self._node_index = None
        else:
            self._node_index = int(value)
        return

    @property
    def target(self):
        """
        The target body for the impact.
        
        Returns
        -------
        Target
        """ 
        return self._target
    
    @target.setter
    def target(self, value):
        if value is None:
            self._target = Target(name="Moon")
            return
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value
        return
    
    @property
    def scale_cls(self):
        """
        The class to use for computing crater scaling relationships.
        
        Returns
        -------
        Type[Scale]
        """ 
        return self._scale_cls

    @scale_cls.setter
    def scale_cls(self, cls):
        if cls is None:
            self._scale_cls = Scale
            return
        if not issubclass(cls, Scale):
            raise ValueError("The class must be a subclass of scale")
        self._scale_cls = cls
        return
        
    @property
    def scale(self):
        """
        An object that defines parameters relavant to scaling the crater size and projectile->crater scaling.
        
        Returns
        -------
        Scale
        """ 
        return self._scale
    
    @scale.setter
    def scale(self, value):
        if value is not None and not isinstance(value, Scale):
            raise TypeError("scale must be an instance of Scale")
        self._scale = value
        return    

    @property
    def rng(self):
        """
        A random number generator instance.
        
        Returns
        -------
        Generator
        """ 
        return self._rng
    
    @rng.setter
    def rng(self, value):
        if not isinstance(value, Generator) and value is not None:
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
        self._rng = value or np.random.default_rng()


class Crater(Impact):
    """
    Represents a crater formed by an impact in the simulation.

    This class models the crater resulting from an impact, including its size,
    shape, depth, and other morphological features.

    Parameters
    ----------
    transient_diameter : FloatLike
        The diameter of the transient crater in m.
    transient_radius : FloatLIke
        The radius of the transient crater in m.    
    morphology_cls : Type[Morphology]
        The class to use for computing the crater morphology.
    **kwargs : Any 
        Additional keyword arguments to pass to the scale and morphology classes.
    """

    def __init__(self, 
                transient_diameter: FloatLike = None,
                transient_radius: FloatLike = None,
                morphology_cls: Type[Morphology] = None, 
                **kwargs: Any):
        """
        Constructor for the Crater class.

        """ 
        self._transient_diameter = None
        self._transient_radius = None
        self._morphology_type = None
        super().__init__(**kwargs)
        
        #  Evaluate and check diameter/radius values 
        values_set = sum(x is not None for x in [self.diameter, transient_diameter, transient_radius])
        if values_set > 1:
            raise ValueError("Only one of diameter, radius, transient_diameter, transient_radius may be set")
        elif values_set == 0:
            raise ValueError("A crater must include one of diameter, radius, transient_diameter, or transient_radius!")                     
        
        # Now call the setters to ensure that all related calculations and checks are performed
        self.transient_diameter = transient_diameter
        self.transient_radius = transient_radius    
        
        self.morphology_cls = morphology_cls
        
        # Make sure kwargs don't get duplicated
        kwargs.pop('crater', None)
        kwargs.pop('target', None)
        kwargs.pop('rng', None)
        self.morphology = self.morphology_cls(crater=self, target=self.target, rng=self.rng, **kwargs)
            
        return

    
    def __repr__(self):
        return (f"Crater(diameter={self.diameter} m, radius={self.radius} m, "
                f"transient_diameter={self.transient_diameter} m, transient_radius={self.transient_radius} m, "
                f"morphology_type={self.morphology_type} "
                f"lon: {self.location[0]}, lat {self.location[1]} age={self.age} My)")
        
    @property
    def diameter(self):
        """
        The diameter of the crater rim in m. Setting diameter automatically sets radius, transient_diameter, and transient_radius.
        
        Returns
        -------
        np.float64
        """
        return self._diameter

    @diameter.setter
    def diameter(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Diameter of crater rim must be finite and positive!")
            self._diameter = np.float64(value)
            self._radius = np.float64(value) / 2
            self._transient_diameter, self.morphology_type = self.scale.final_to_transient(value)
            self._transient_radius = self._transient_diameter / 2
    
    @property
    def radius(self):
        """
        The radius of the crater rim in m. Setting radius automatically sets diameter, transient_diameter, and transient_radius.
        
        Returns
        -------
        np.float64
        """        
        return self._radius

    @radius.setter
    def radius(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Radius of crater rim must be finite and positive!")            
            self._radius = np.float64(value)
            self._diameter = np.float64(value) * 2
            self._transient_diameter, self.morphology_type = self.scale.final_to_transient(value)
            self._transient_radius = self._transient_diameter / 2    
            
    @property
    def transient_diameter(self):
        """
        The diameter of the transient crater in m. Setting transint diameter automatically sets transient radius, diameter, and radius.
        
        Returns
        -------
        np.float64
        """        
        return self._transient_diameter

    @transient_diameter.setter
    def transient_diameter(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Diameter of transient crater must be finite and positive!")
            self._transient_diameter = np.float64(value)
            self._transient_radius = np.float64(value) / 2
            self._diameter,self.morphology_type = self.scale.transient_to_final(value)
            self._radius = self._diameter / 2
        return
    
    @property
    def transient_radius(self):
        """
        The radius the transient crater in m. Setting transint radius automatically sets transient diameter, diameter, and radius.
        
        Returns
        -------
        np.float64
        """            
        return self._transient_radius

    @transient_radius.setter
    def transient_radius(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Radius of transient crater must be finite and positive!")            
            self._transient_radius = np.float64(value)
            self._transient_diameter = np.float64(value) * 2
            self._diameter,self.morphology_type = self.scale.transient_to_final(value)
            self._radius = self._diameter / 2
    

    @property
    def morphology_cls(self):
        """
        The class to use for computing the crater morphology.
        
        Returns
        -------
        Type[Morphology]
        """ 
        return self._morphology_cls

    @morphology_cls.setter
    def morphology_cls(self, cls):
        if cls is None:
            self._morphology_cls = Morphology
        elif not issubclass(cls, Morphology):
            raise ValueError("The class must be a subclass of Morphology")
        else:
            self._morphology_cls = cls
        
    @property
    def morphology(self):
        """
        An object that defines parameters relavant to describing the morphology of the crater.
        
        Returns
        -------
        Morphology
        """ 
        return self._morphology
    
    @morphology.setter
    def morphology(self, value):
        if not isinstance(value, Morphology):
            raise TypeError("morphology must be an instance of Morphology")
        self._morphology = value
        return 

    @property
    def morphology_type(self):
        """
        The type of morphology to use for the crater.
        
        Returns
        -------
        str
        """ 
        return self._morphology_type
    
    @morphology_type.setter
    def morphology_type(self, value):
        if not isinstance(value, str):
            raise TypeError("morphology_type must be a string")
        self._morphology_type = value
        return


class Projectile(Impact):
    """
    Represents the projectile body in the crater simulation.

    This class defines the properties of the impacting object, such as its size,
    velocity, material, and angle of impact.
    
    Parameters
    ----------
    mass : float
        The mass of the projectile in kg.
    density : float
        The mass density of the projectile in kg/m**3.
    mean_velocity : float
        The mean impact velocity of the projectile population in m/s. The velocity will be computed from the get_random_velocity
        function in the montecarlo module.  `mean_velocity` cannot be used with `velocity` or `vertical_velocity`.
    velocity : float
        The velocity of the projectile upon impact, in m/s.
    vertical_velocity : float
        The vertical component of the projectile velocity upon impact, in m/s.
    angle : float
        The angle of impact, in degrees.
    direction: float
        The direction of impact in degrees relative to north (the bearing on the surface)
    **kwargs : Any
    
    Notes
    -----
    The Projectile class is initialized with at least one of diameter, radius, or mass, but only two of mass, density, or
    radius or two of mass, density, and diameter, and only two of velocity, vertical_velocity, or angle.  If density cannot be
    determined from the inputs, the density of the target body is used as a default. If the velocity cannot be determined from 
    the inputs, the mean impact velocity of the target is used as an input to the get_random_velocity function. If the angle 
    cannot be determined from the inputs, a random angle is generated using the get_random_impact_angle function.
    """
    
    def __init__(self, 
                mass: FloatLike = None,
                density: FloatLike = None,
                mean_velocity: FloatLike = None,
                velocity: FloatLike = None,
                vertical_velocity: FloatLike = None,
                angle: FloatLike = None,
                direction: FloatLike = None,
                **kwargs):
        """
        
        Constructor for the Projectile class.
        

        """ 

        # ensure that all related calculations and checks are performed
        self._density = density
        self._mass = mass
        self._mean_velocity = None 
        self._velocity = None
        self._vertical_velocity = None
        self._angle = None
        self._direction = direction
        
        super().__init__(**kwargs)
        values_set = sum(x is not None for x in [self.diameter, self.radius, mass])
        # Evaluate and check mass/density/radius inputs 
        if values_set == 0:
            raise ValueError("A projectile must include one of diameter, radius, or mass!")
        values_set = sum(x is not None for x in [mass, density, self.radius])
        if values_set > 2:
            raise ValueError("Only two of mass, density, and radius may be set")
        values_set = sum(x is not None for x in [mass, density, self.diameter])
        if values_set > 2:
            raise ValueError("Only two of mass, density, and diameter may be set")        
        
        if self.mass is None or self.radius is None: # Default to target density if we are given no way to figure it out
            self.density = self.scale.material.density 
           
        if self.radius is None: 
            self._update_volume_based_properties()
        
        if self.mass is None:
            self._update_mass()
        
        # Evaluate and check velocity/impact angle inputs
        if mean_velocity is not None:
            if velocity is not None or vertical_velocity is not None:
                raise ValueError("mean_velocity cannot be used with velocity or vertical_velocity")
            self.mean_velocity = mean_velocity
        else:    
            values_set = sum(x is not None for x in [velocity, vertical_velocity, angle])
            if values_set > 2:
                raise ValueError("Only two of velocity, vertical_velocity, angle may be set")        
            self._velocity = velocity    
            self._angle = angle    
            self._vertical_velocity = vertical_velocity
            self._initialize_velocities()
            
         
        return

    def __repr__(self):
        return (f"Projectile(diameter={self.diameter} m, radius={self.radius} m, "
                f"mass={self.mass} kg, density={self.density} kg/m^3, "
                f"velocity={self.velocity} m/s, angle={self.angle} deg, "
                f"vertical_velocity={self.vertical_velocity} m/s, direction={self.direction} deg,"
                f"lon: {self.location[0]}, lat {self.location[1]} age={self.age} My)")


    @property
    def diameter(self):
        """
        The diameter of the projectile in m. Setting the diameter automatically sets the radius. Mass and density are updated accordingly.
        
        Returns
        -------
        np.float64 
        """       
        return self._diameter

    @diameter.setter
    def diameter(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Diameter of projectile must be finite and positive!")
            self._diameter = np.float64(value)
            self._radius = np.float64(value) / 2
            self._update_mass()
        return

    @property
    def radius(self):
        """
        The radius of the projectile in m. Setting the radius automatically sets the diameter. Mass and density are updated accordingly.
        
        Returns
        -------
        np.float64 
        """
        return self._radius

    @radius.setter
    def radius(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Radius of projectile must be finite and positive!")            
            self._radius = np.float64(value)
            self._diameter = np.float64(value) * 2
            self._update_mass()
        return

    @property
    def mass(self):
        """
        The mass of the projectile in kg. Setting the mass automatically updates the diameter and radius from a given density.
        
        Returns
        -------
        np.float64 
        """
        return self._mass

    @mass.setter
    def mass(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Mass of projectile must be finite and positive!")            
            self._mass = np.float64(value)
            self._update_volume_based_properties()
        return

    @property
    def density(self):
        """
        The mass density of the projectile in kg/m**3. Setting the density automatically updates the mass from a given radius.
        
        Returns
        -------
        np.float64 
        """
        return self._density

    @density.setter
    def density(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Density of projectile must be finite and positive!")                 
            self._density = np.float64(value)
            self._update_mass()

    def _update_mass(self):
        if self._density is not None and self._radius is not None:
            self._mass = 4.0/3.0 * np.pi * self._radius**3 * self._density

    def _update_volume_based_properties(self):
        if self._mass is not None and self._density is not None:
            volume = self._mass / self._density
            self._radius = ((3.0 * volume) / (4.0 * np.pi))**(1.0/3.0)
            self._diameter = self._radius * 2
            
    @property
    def mean_velocity(self):
        """
        The mean impact velocity of the projectile population in m/s. The velocity will be computed from the get_random_velocity
        function in the montecarlo module.  `mean_velocity` cannot be used with `velocity` or `vertical_velocity`.
        
        Returns
        -------
        np.float64 
        """
        return self._mean_velocity
   
    @mean_velocity.setter 
    def mean_velocity(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Mean velocity of projectile must be finite and positive!")            
            self._velocity = None
            self._vertical_velocity = None
            self._mean_velocity = np.float64(value)
            self._initialize_velocities()
        return 
    

    @property
    def velocity(self):
        """
        The velocity of the projectile upon impact, in m/s. Setting the velocity automatically updates the vertical velocity from a given angle.
        
        Returns
        -------
        np.float64 
        """
        if self._velocity is None and self._vertical_velocity is not None and self._angle is not None:
            self._velocity = np.float64(self._vertical_velocity / np.sin(np.deg2rad(self._angle)))
        return self._velocity

    @velocity.setter
    def velocity(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Velocity of projectile must be finite and positive!")        
            self._velocity = np.float64(value)
            
            if value is not None and self._angle is not None:
                self._vertical_velocity = np.float64(value) * np.sin(np.deg2rad(self._angle))
            self._initialize_velocities()

    @property
    def vertical_velocity(self):
        """
        The vertical component of the projectile velocity upon impact, in m/s. Setting the vertical velocity automatically updates the velocity from a given angle. 
        The vertical velocity must be less than or equal to the velocity.
        
        Returns
        -------
        np.float64 
        """
        if self._vertical_velocity is None and self._velocity is not None and self._angle is not None:
            self._vertical_velocity = np.float64(self._velocity * np.sin(np.deg2rad(self._angle)))
        return self._vertical_velocity

    @vertical_velocity.setter
    def vertical_velocity(self, value):
        if value is not None:
            if value <= 0.0:
                raise ValueError("Vertical velocity of projectile must be finite and positive!")        
            self._vertical_velocity = np.float64(value)
        
            # Update velocity only if angle is already set
            if self._angle is not None and value is not None:
                try:
                    self._velocity = np.float64(value) / np.sin(np.deg2rad(self._angle))
                except ValueError:
                    raise ValueError(f"Invalid vertical velocity value {value} for a given angle value {self._angle}!")
            self._initialize_velocities()

    @property
    def angle(self):
        """
        The angle of impact, in degrees. Setting the angle automatically updates the vertical velocity from a given velocity.
        
        Returns
        -------
        np.float64
        """
        return self._angle

    @angle.setter
    def angle(self, value):
        if value is not None:
            if value < 0.0 or value > 90.0:
                raise ValueError("Impact angle of projectile must be between 0 and 90 degrees")
            self._angle = np.float64(value)
            # Update vertical_velocity only if velocity is already set
            if self._velocity is not None:
                self._vertical_velocity = self._velocity * np.sin(np.deg2rad(self._angle))
            self._initialize_velocities()

    def _initialize_velocities(self):
        """
        Initialize the impact velocity and angle of the projectile. 
        """ 
        if self._velocity is None:
            if self._mean_velocity is None:
                raise ValueError("No mean impact velocity is defined. Projectiles will not be computed.")
        
            # Check if the impact velocity is greater than the escape velocity. If so, be sure to add the escape velocity to the mean impact velocity in quadrature
            # Otherwise, just use the mean velocity value
            if self._mean_velocity > self.target.escape_velocity:
                vencounter_mean = np.sqrt(self._mean_velocity**2 - self.target.escape_velocity**2)
                vencounter = mc.get_random_velocity(vencounter_mean, rng=self.rng)
                self._velocity = np.sqrt(vencounter**2 + self.target.escape_velocity**2, dtype=np.float64)
            else: 
                # Be sure not to generate velocities above escape in this scenario
                while True:  
                    self._velocity = mc.get_random_velocity(self._mean_velocity, rng=self.rng)
                    if self._velocity < self.target.escape_velocity:
                        break

        if self._angle is None:
            self._angle = mc.get_random_impact_angle(rng=self.rng)
                
        if self._direction is None:
            self._direction = np.float64(self.rng.uniform(0.0, 360.0))
        return

    @property
    def direction(self):
        """
        The direction of impact in degrees relative to north (the bearing on the surface)
        
        Returns
        -------
        np.float64
        """
        return self._direction
    
    @direction.setter
    def direction(self, value):
        if value is not None:
            if value < 0.0 or value >= 360.0:
                raise ValueError("Direction of impact must be between 0 and 360 degrees")
            self._direction = np.float64(value)
        return


import unittest
from cratermaker import NeukumProduction, Production
import numpy as np

class TestProduction(unittest.TestCase):
    
    def test_production_N_to_time(self):

        plaw = Production(impact_velocity_model="Mercury_MBA")
        neukum = NeukumProduction(model="Moon")

        for age_orig in np.linspace(0,4500,num=10):
            D = np.logspace(-1,3,num=1000)
            N = neukum.function(diameter=D,age=age_orig)
            age_new = neukum.function_inverse(cumulative_number_density=N,diameter=D)
            np.testing.assert_array_almost_equal(age_orig,age_new,decimal=2)
            
            N = plaw.function(diameter=D,age=age_orig)
            age_new = plaw.function_inverse(cumulative_number_density=N,diameter=D)
            np.testing.assert_array_almost_equal(age_orig,age_new,decimal=2)
            
    def test_sample_arguments(self):
        production = [Production(impact_velocity_model="MBA_MBA"), NeukumProduction()]
        for prod in production:

            # Test valid arguments
            try:
                prod.sample(age=500.0, diameter_range=(10.0, 100.0), area=1e6)
            except ValueError:
                self.fail("sample() raised ValueError unexpectedly with valid arguments!")

            # Test providing both age and cumulative_number
            with self.assertRaises(ValueError):
                prod.sample(age=500, diameter_number=(100, 10), diameter_range=(10, 100))
               
            # Test providing both age_end and diameter_number_end
            with self.assertRaises(ValueError):
                prod.sample(age=1500, age_end=1000, diameter_number_end=(100, 10), diameter_range=(10, 100))

            # Test missing diameter_range
            with self.assertRaises(ValueError):
                prod.sample(age=500)

            # Test diameter_range with invalid length
            with self.assertRaises(ValueError):
                prod.sample(age=500, diameter_range=(10,), area=1000)

            # Test diameter_range with non-positive minimum diameter
            with self.assertRaises(ValueError):
                prod.sample(age=500, diameter_range=(-1, 100), area=1000)

            # Test diameter_range with max diameter less than min diameter
            with self.assertRaises(ValueError):
                prod.sample(age=500, diameter_range=(100, 10), area=1000)

            # Test invalid diameter_number length
            with self.assertRaises(ValueError):
                prod.sample(diameter_number=(100,), diameter_range=(10, 100))

            # Test negative value in diameter_number
            with self.assertRaises(ValueError):
                prod.sample(diameter_number=(-100, 10), diameter_range=(10, 100))

            # Test negative area
            with self.assertRaises(ValueError):
                prod.sample(age=500, diameter_range=(10, 100), area=-1000)
                
            # Test non-scalar age
            with self.assertRaises(ValueError):
                prod.sample(age=[500, 1000], diameter_range=(10, 100))
            
            # This checks if a ValueError is raised for an invalid length of diameter_number_end.    
            with self.assertRaises(ValueError):
                prod.sample(diameter_number_end=(100,), diameter_range=(10, 100))
           
            #Testing when a negative value is provided in diameter_number_end. 
            with self.assertRaises(ValueError):
                prod.sample(diameter_number_end=(-100, 10), diameter_range=(10, 100))

    def test_small_sample(self):
        neukum = NeukumProduction(model="Moon")
        # Test that the number of craters is zero when the age and area are ridiculously tiny
        diameter, age = neukum.sample(age=1e-6, diameter_range=(300e3, 1000e3), area=1e-6)
        self.assertEqual(diameter.size, 0)
        self.assertEqual(age.size, 0) 
        
if __name__ == '__main__':
    unittest.main()


# cython: language_level=3, c_string_type=unicode, c_string_encoding=ascii
cimport cython
cimport numpy as cnp
import numpy as np
from libc.stdlib cimport malloc, free

cdef extern from "realistic.h":
    ctypedef struct PerlinArguments:
        double damp
        double damp0
        double damp_scale
        double freq
        double gain
        double gain0
        double lacunarity
        double noise_height
        double pers
        double slope
        double warp
        double warp0

    double bind_realistic_perlin_noise_one(const char *model, double x, double y, double z, int num_octaves, double *anchor, double damp, double damp0, double damp_scale, double freq, double gain, double gain0, double lacunarity, double noise_height, double pers, double slope, double warp, double warp0)

    void bind_realistic_perlin_noise_all(const char *model, double *x, double *y, double *z, int num_elements, int num_octaves, double *anchor, double damp, double damp0, double damp_scale, double freq, double gain, double gain0, double lacunarity, double noise_height, double pers, double slope, double warp, double warp0, double *noise)

cdef void _to_fortran_2D_double_array(cnp.ndarray[cnp.float64_t, ndim=2] src, double* dest, int rows, int cols):
    """
    Copies data from a 2D NumPy array to a 2D Fortran-style array.

    Parameters
    ----------
    src : cnp.ndarray
        The source NumPy array with 2 dimensions.
    dest : double*
        The pointer to the destination array in memory which should be already allocated.
    rows : int
        The number of rows in the destination array.
    cols : int
        The number of columns in the destination array.
    """
    # Assuming dest is already allocated with the correct size
    # NumPy arrays are row-major, while Fortran arrays are column-major,
    # so we need to transpose the indices when copying.
    for i in range(rows):
        for j in range(cols):
            dest[j * rows + i] = src[i, j]  # Transpose the index for Fortran's column-major order


def apply_noise(str model, 
                cnp.ndarray[cnp.float64_t, ndim=1] x_array, 
                cnp.ndarray[cnp.float64_t, ndim=1] y_array, 
                cnp.ndarray[cnp.float64_t, ndim=1] z_array, 
                int num_octaves, 
                cnp.ndarray[cnp.float64_t, ndim=2] anchor, 
                **kwargs):
    """
    Applies Perlin noise based on the specified model and parameters.

    Parameters
    ----------
    model: str  
        Name of the turbulence model.
    x_array, y_array, z_array: ndarray (N,)  
        Cartesian coordinates for noise evaluation.
    num_octaves: int 
        Number of octaves for the noise function.
    anchor: ndarray (num_octaves, 3)
        array of spatial anchor points for each noise octave.
    kwargs: dict 
        Additional model-specific parameters.

    Returns
    -------
    ndarray(N,)
        computed noise value.

    Raises
    ------
    ValueError - If required parameters are missing, arrays are mismatched, or if an invalid model is specified.
    """


    if not (x_array.size == y_array.size == z_array.size):
        raise ValueError("x, y, and z arrays must have the same length")

    cdef int num_elements = x_array.size

    # Ensure memory-contiguous numpy array
    anchor = np.ascontiguousarray(anchor, dtype=np.float64)
    x_array = np.ascontiguousarray(x_array, dtype=np.float64)
    y_array = np.ascontiguousarray(y_array, dtype=np.float64)
    z_array = np.ascontiguousarray(z_array, dtype=np.float64)
    noise_array = np.empty(num_elements, dtype=np.float64)

    # Make memory view of the numpy arrays
    cdef cnp.float64_t[::1] x = x_array
    cdef cnp.float64_t[::1] y = y_array
    cdef cnp.float64_t[::1] z = z_array
    cdef cnp.float64_t[::1] noise = noise_array

    # Convert numpy 2D array to double**
    cdef double* f_anchor = <double*> malloc(3 * num_octaves * sizeof(double))
    _to_fortran_2D_double_array(anchor, f_anchor, num_octaves, 3)

    model = model.lower()
    kw = {
        'damp' : 0.0, 
        'damp0' : 0.0, 
        'damp_scale' : 0.0, 
        'freq' : 0.0, 
        'gain' : 0.0, 
        'gain0' : 0.0, 
        'lacunarity' : 0.0, 
        'noise_height' : 0.0, 
        'pers' : 0.0, 
        'slope' : 0.0, 
        'warp' : 0.0, 
        'warp0' : 0.0
    }
    required_kwargs = {
        'turbulence' : ['noise_height', 'freq', 'pers'],
        'billowed'   : ['noise_height', 'freq', 'pers'],
        'plaw'       : ['noise_height', 'freq', 'pers', 'slope'],
        'ridged'     : ['noise_height', 'freq', 'pers'],
        'swiss'      : ['lacunarity', 'gain', 'warp'],
        'jordan'     : ['lacunarity', 'gain0', 'gain', 'warp0', 'warp', 'damp0', 'damp', 'damp_scale'],
    }

    if model in required_kwargs:
        # If it is a valid model, make sure we've passed all of the required arguments
        if set(required_kwargs[model]).issubset(kwargs.keys()):
            # iterate over required_kwargs[model] for all keys and set the value of kw[key]to be given by kwargs[key]
            for arg in required_kwargs[model]:
                kw[arg] = kwargs[arg]

            bind_realistic_perlin_noise_all(model, &x[0], &y[0], &z[0], num_elements, num_octaves, f_anchor, kw['damp'], kw['damp0'], kw['damp_scale'], kw['freq'], kw['gain'], kw['gain0'], kw['lacunarity'], kw['noise_height'], kw['pers'], kw['slope'], kw['warp'], kw['warp0'], &noise[0] )
        else:
            missing_args = set(required_kwargs[model]) - kwargs.keys()
            raise ValueError(f"The {model} model requires the following missing keywords: {', '.join(missing_args)}")
    else:
        raise ValueError(f"{model} is an invalid model. Valid options are {', '.join(required_kwargs.keys())}")

    # Free the allocated memory
    free(f_anchor)

    return noise  # Return the noise value

def realistic_crater(cnp.ndarray[cnp.float64_t, ndim=1] radial_distance, 
                cnp.ndarray[cnp.float64_t, ndim=1] initial_bearing,
                cnp.ndarray[cnp.float64_t, ndim=1] reference_elevation_array, 
                cnp.ndarray[cnp.float64_t, ndim=1] rim_1D_psd,
                cnp.ndarray[cnp.float64_t, ndim=1] floor_1D_psd,
                cnp.ndarray[cnp.float64_t, ndim=1] ejecta_1D_psd,
                cnp.ndarray[cnp.float64_t, ndim=2] floor_2D_psd,
                cnp.ndarray[cnp.float64_t, ndim=2] ejecta_2D_psd,
                cnp.ndarray[cnp.float64_t, ndim=2] wall_2D_psd,
                cnp.float64_t crater_diameter):
    """
    Generate an ejecta profile.

    Parameters
    ----------
    radial_distance : ndarray(N,)
        Radial distance from the center of the crater.
    initial_bearing : ndarray(N,)
        Initial bearing from the denter of the crater in radians.
    crater_diameter : float
        Diameter of the crater.


    Returns
    -------
    ndarray(N,)
        computed elevation values

    Raises
    ------
    ValueError - If required parameters are missing, arrays are mismatched, or if an invalid model is specified.
    """

    cdef int num_elements = radial_distance.size



import numpy as np
from typing import Dict, Optional, Any
from ..utils.general_utils import set_properties, create_catalogue, check_properties
from ..utils.custom_types import FloatLike

class Target:
    """
    Represents the target body in a crater simulation.

    This class encapsulates the properties of the target that is impacted, including
    its material composition, size, and other relevant physical characteristics.
    """
       
    config_ignore = ['catalogue','material']  # Instance variables to ignore when saving to file
    def __init__(self, 
                 name: str, 
                 radius: FloatLike | None = None, 
                 diameter: FloatLike | None = None, 
                 gravity: FloatLike | None = None, 
                 transition_scale_type: str | None = None,
                 material_name: str | None = None,                  
                 catalogue: Dict[str, Dict[str, FloatLike]] | None = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the target object, setting properties from the provided arguments,
        and creating a catalogue of known solar system targets if not provided.

        Parameters
        ----------
        name : str or None
            Name of the target body.
        radius : FloatLike or None
            Radius of the target body in meters.
        diameter : FloatLike or None
            Diameter of the target body in meters.
        gravity : FloatLike or None
            Surface gravity of the target body in m/s^2.
        material_name : str or None
            Name of the material composition of the target body.            
        transition_scale_type : str or None
            Simple-to-complex transition scaling to use for the surface (either "silicate" or "ice").
        catalogue : dict or None
            Optional input of catalogue solar system targets to replace the built-in catalogue.
        **kwargs : Any
            Additional keyword argumments that could be set by the user.
        """    

        # Set the attributes for this class
        self._name = name
        self._gravity = None
        self._transition_scale_type = None
        self._material_name = None
        
        # ensure that only either diamter of radius is passed
        values_set = sum(x is not None for x in [diameter, radius])
        if values_set > 1:
            raise ValueError("Only one of diameter, radius may be set")        
        
        self._radius = None
        self._diameter = None
        self._catalogue = None
        self.catalogue = catalogue 
            
        # Set properties for the Target object based on the arguments passed to the function
        self.set_properties(name=name, 
                            radius=radius, 
                            diameter=diameter, 
                            gravity=gravity, 
                            material_name=material_name,
                            transition_scale_type=transition_scale_type, 
                            catalogue=self.catalogue,
                            **kwargs)

        check_properties(self) 
        return
    
    @property
    def catalogue(self):
        """
        A nested dictionary containing a catalogue of known solar system targets to use for the simulation. 
        
        Returns
        -------
        Dict[str, Dict[str, FloatLike]] or None
            A catalogue of known solar system targets to use for the simulation.
        """
        return self._catalogue

    @catalogue.setter
    def catalogue(self, value):
        
        if not isinstance(value, dict) and value is not None:
            raise TypeError("catalogue must be a dict or None")
        
        # Define some built-in catalogue values for known solar system targets of interest
        gEarth = np.float64(9.80665) # 1 g in SI units
        body_properties = [
            "name",    "radius",   "gravity",      "material_name", "transition_scale_type"
        ]
        body_values = [
            ("Mercury", 2440.0e3,  0.377 * gEarth, "Soft Rock", "silicate"),
            ("Venus",   6051.84e3, 0.905 * gEarth, "Hard Rock", "silicate"),
            ("Earth",   6371.01e3, 1.000 * gEarth, "Wet Soil" , "silicate"),
            ("Moon",    1737.53e3, 0.1657* gEarth, "Soft Rock", "silicate"),
            ("Mars",    3389.92e3, 0.379 * gEarth, "Soft Rock", "silicate"),
            ("Ceres",   469.7e3,   0.029 * gEarth, "Ice"      , "ice"),
            ("Vesta",   262.7e3,   0.025 * gEarth, "Soft Rock", "silicate"),
        ]
        
        if value is None: 
            self._catalogue = create_catalogue(body_properties, body_values)
        else:
            self._catalogue = value    
    

    def set_properties(self, **kwargs):
        """
        Set properties of the current object based on the provided keyword arguments.

        This function is a utility to update the properties of the current object. The actual implementation of the 
        property setting is handled by the `util.set_properties` method.

        Parameters
        ----------
        **kwargs : dict
            A dictionary of keyword arguments that represent the properties to be set on the current object.

        Returns
        -------
        None
            The function does not return a value.
        """         
        set_properties(self,**kwargs)
        return

    
    @property
    def escape_velocity(self):
        """
        Calculate the escape velocity for the target body.

        Returns
        -------
        np.float64
            Escape velocity in m/s.
        """        
        return np.sqrt(2 * self.radius * self.gravity)

    
    @property
    def name(self):
        """
        The name of the target body.
        
        Returns
        -------
        str 
        """
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be a string or None")
        self._name = value

       
    @property
    def radius(self) -> Optional[float]:
        """
        Radius of the target body in m.
        
        Returns
        -------
        np.float64 
        """        
        return self._radius

    @radius.setter
    def radius(self, value: float):
        if not isinstance(value, float) or value <= 0:
            raise ValueError("Radius must be a positive number.")
        self._radius = np.float64(value)
        self._diameter = np.float64(value) * 2

    @property
    def diameter(self) -> Optional[float]:
        """
        Diameter of the target body in m.
        
        Returns
        -------
        np.float64 
        """                
        return self._diameter

    @diameter.setter
    def diameter(self, value: float):
        if not isinstance(value, float) or value <= 0:
            raise ValueError("Diameter must be a positive number.")
        self._diameter = np.float64(value)
        self._radius = np.float64(value) / 2

    @property
    def gravity(self):
        """
        Surface gravity of the target body in m/s^2.
        
        Returns
        -------
        np.float64
        """
        return self._gravity

    @gravity.setter
    def gravity(self, value):
        if value is not None and not isinstance(value, FloatLike):
            raise TypeError("gravity must be a numeric value or None")
        self._gravity = np.float64(value)

    @property
    def material_name(self):
        """
        The name of the material composition of the target body.
        
        Returns
        -------
        str 
        """
        return self._material_name

    @material_name.setter
    def material_name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("material_name must be a string or None")
        self._material_name = value
        
    @property
    def transition_scale_type(self):
        """
        The type of simple-to-complex transition scaling used for the surface, either 'silicate' or 'ice'.
        
        Returns
        -------
        str 
        """
        return self._transition_scale_type

    @transition_scale_type.setter
    def transition_scale_type(self, value):
        valid_types = ["silicate", "ice"]
        if value not in valid_types and value is not None:
            raise ValueError(f"Invalid transition_scale_type: {value}. Must be one of {valid_types} or None")
        self._transition_scale_type = value



import numpy as np
from numpy.random import Generator
from typing import Tuple, Any, Dict
from scipy.optimize import root_scalar
from .target import Target
from ..utils.custom_types import FloatLike
from ..utils import montecarlo as mc
from ..utils.general_utils import set_properties, create_catalogue, check_properties

class Material:
    """
    Represents the material properties relevant to the crater simulation.

    This class defines various physical properties of the material involved in the cratering process.

    """
    config_ignore = ['catalogue']  # Instance variables to ignore when saving to file
    def __init__(self,
                 name: str | None = None,
                 K1: FloatLike | None = None,
                 mu: FloatLike | None = None,
                 Ybar: FloatLike | None = None,
                 density: FloatLike | None = None,
                 catalogue: Dict[str, Dict[str, FloatLike]] | None = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the target object, setting properties from the provided arguments,
        and creating a catalogue of known solar system targets if not provided.
        
        Parameters
        ----------
        name : str
            The name of the material. If the material is matched to one that is present in the catalogue, the rest of the properties 
            will be retrieved for it unless specified. If the name is not known from the catalogue, then all other properties must 
            be supplied and in order to build a custom material.
        K1 : FloatLike
            Variable used in crater scaling (see _[1])
        mu : FloatLike
            Variable used in crater scaling (see _[1])
        Ybar : FloatLike
            The strength of the material, (Pa)
        density : FloatLike
            Volumentric density of material, (kg/m^3)
        catalogue : Dict[str, Dict[str, FloatLike]]
            An optional dictionary containing a catalogue of known materials to use for the simulation. The catalogue should be 
            constructed using a nested dictionary, where the first level of keys are the names of the materials, and the second level
            are the corresponding property names (K1, mu, Ybar, density). 
            If not provided, a default catalogue will be used.
        **kwargs : Any
            Additional keyword argumments that could be set by the user.
            
        Notes
        -----
        The material properties defined here include crater scaling relationship values that were used in CTEM from Richardson (2009) [1]_.  These values used are from Holsapple (1993) [2]_ and Kraus et al. (2011) [3]_.
        
        References
        ----------
        .. [1] Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697–715. https://doi.org/10.1016/j.icarus.2009.07.029
        .. [2] Holsapple, K.A., 1993. The scaling of impact processes in planetary sciences 21, 333–373. https://doi.org/10.1146/annurev.ea.21.050193.002001
        .. [3] Kraus, R.G., Senft, L.E., Stewart, S.T., 2011. Impacts onto H2O ice: Scaling laws for melting, vaporization, excavation, and final crater size. Icarus 214, 724–738. https://doi.org/10.1016/j.icarus.2011.05.016        

        """ 
        # Set the attributes for this class
        self._name = name
        self._K1 = None
        self._mu = None
        self._Ybar = None
        self._density = None
        self._catalogue = None
       
        self.catalogue = catalogue
        
        # Set properties for the Material object based on the arguments passed to the function
        self.set_properties(name=name,
                            K1=K1,
                            mu=mu,
                            Ybar=Ybar,
                            density=density,
                            catalogue=self.catalogue,
                            **kwargs) 
        
        # Check to make sure all required properties are set 
        check_properties(self)
        
        return    

    @property
    def catalogue(self):
        """
        A nested dictionary containing a catalogue of known materials to use for the simulation. 
        
        Returns
        -------
        Dict[str, Dict[str, FloatLike]] or None
            A catalogue of known materials to use for the simulation.
        """
        return self._catalogue
    
    @catalogue.setter
    def catalogue(self, value):
                      
        if not isinstance(value, dict) and value is not None:
            raise TypeError("catalogue must be a dict or None") 
             
        # Define some default crater scaling relationship terms (see Richardson 2009, Table 1, and Kraus et al. 2011 for Ice) 
        material_properties = [
            "name",       "K1",     "mu",   "Ybar",     "density" 
        ]
        material_values = [
            ("Water",     2.30,     0.55,   0.0,        1000.0),
            ("Sand",      0.24,     0.41,   0.0,        1750.0),
            ("Dry Soil",  0.24,     0.41,   0.18e6,     1500.0),
            ("Wet Soil",  0.20,     0.55,   1.14e6,     2000.0),
            ("Soft Rock", 0.20,     0.55,   7.60e6,     2250.0),
            ("Hard Rock", 0.20,     0.55,   18.0e6,     2500.0),
            ("Ice",       15.625,   0.48,   0.0,        900.0), 
        ]        
       
        if value is None: 
            self._catalogue = create_catalogue(material_properties, material_values)
        else:
            self._catalogue = value    

    @property
    def name(self):
        """
        The name of the material.
        
        Returns
        -------
        str 
            Name of the material.
        """
        return self._name
    
    @name.setter
    def name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("name must be a string or None")
        self._name = value
        
    @property
    def K1(self):
        """
        K1 crater scaling relationship term. 
        
        Returns
        -------
        np.float64 
        
        References
        ----------
        Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697–715. https://doi.org/10.1016/j.icarus.2009.07.029 
        """
        return self._K1
    
    @K1.setter
    def K1(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("K1 must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("K1 must be a positive number")
        self._K1 = np.float64(value)
        
    @property
    def mu(self):
        """
        mu crater scaling relationship term.
        
        Returns
        -------
        np.float64 
        
        References
        ----------
        Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697–715. https://doi.org/10.1016/j.icarus.2009.07.029 
        """
        return self._mu
    
    @mu.setter
    def mu(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("mu must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("mu must be a positive number")
        self._mu = np.float64(value)
        
    @property
    def Ybar(self):
        """
        The strength of the material in Pa.
        
        Returns
        -------
        np.float64 
            
                    
        References
        ----------
        Richardson, J.E., 2009. Cratering saturation and equilibrium: A new model looks at an old problem. Icarus 204, 697–715. https://doi.org/10.1016/j.icarus.2009.07.029 
        """
        return self._Ybar
    
    @Ybar.setter
    def Ybar(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("Ybar must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("Ybar must be a positive number")
        self._Ybar = np.float64(value)
        
    @property
    def density(self):
        """
            Volumentric density of material in kg/m^3.
        
        Returns
        -------
        np.float64 
        """
        return self._density
    
    @density.setter
    def density(self, value):
        if not isinstance(value, FloatLike) and value is not None:
            raise TypeError("density must be a numeric value or None")
        if value is not None and value < 0:
            raise ValueError("density must be a positive number")
        self._density = np.float64(value)
        
    
    def set_properties(self, **kwargs):
        """
        Set properties of the current object based on the provided keyword arguments.

        This function is a utility to update the properties of the current object. The actual implementation of the 
        property setting is handled by the `util.set_properties` method.

        Parameters
        ----------
        **kwargs : dict
            A dictionary of keyword arguments that represent the properties to be set on the current object.

        Returns
        -------
        None
            The function does not return a value.
        """         
        set_properties(self,**kwargs)
        return
    
      
class Scale():
    """
    An operations class for computing the scaling relationships between impactors and craters.

    This class encapsulates the logic for converting between projectile properties and crater properties, 
    as well as determining crater morphology based on size and target properties.

    """

    def __init__(self, 
                 target: Target,
                 material: Material | None = None,
                 material_name: str | None = None,                  
                 rng: Generator | None = None, 
                 **kwargs):
        """
        Create an operations class for computing the scaling relationships between impactors and craters.
        
        Parameters
        ----------
        target : Target
            The target body for the impact simulation.
        material : Material or None
            Material composition of the target body. This will override the material supplied by name given by the target object. 
            Only one of material or material_name can be provided.
        material_name : str or None
            Name of the material composition of the target body. This will override the material supplied by name given by the
            target object. Only one of material or material_name can be provided.
        rng : Generator, optional
            A random number generator instance. If not provided, the default numpy RNG will be used. 
        """
        
        self.rng = rng
        self.target = target
        self._material = None
        self._material_name = None
        
        if material is not None and material_name is not None:
            raise ValueError("Only one of material or material_name can be provided")
        
        if self.target.material_name is None:
            if material_name is None or material is None:
                raise ValueError("The target does not provide a material name. Either material or material_name must be provided")
       
        if material is not None:
            self.material = material
        elif material_name is not None: 
            self.material_name = material_name
            self.material = Material(name=self.material_name)        
        else:
            self.material_name = self.target.material_name
            self.material = Material(name=self.material_name)
        
        # Initialize additional attributes for simple->complex transition scale factors. These are set to None here just for clarity
        self.transition_diameter = None
        self.transition_nominal = None
        self.simple_enlargement_factor = None
        self.complex_enlargement_factor = None
        self.final_exp = None

        # Initialize transition factors
        self._compute_simple_to_complex_transition_factors() 
        return
        
        
    def _compute_simple_to_complex_transition_factors(self):
        """
        Computes and sets the internal attributes for transition factors between simple and complex craters.
        """    
        # These terms are used to compute the ratio of the transient crater to simple crater size       
        simple_enlargement_mean = 0.84 # See Melosh (1989) pg. 129 just following eq. 8.2.1
        simple_enlargement_std = 0.04 # Just using Pike (1980) fig. 9 the crater depth varies by about the same amount on either side of the transition so this is a reasonable assumption
        
        # These terms are used in the exponent in the final rim radius/ simple crater radius vs  final radius / transition radius relationship
        # See Holsapple (1993) eq. 28
        final_exp_mean = 0.079    
        final_exp_std = 0.0001 # We add noise because this is nature and nature messy
        complex_enlargement_factor = 1.02
    
        # These terms are used to compute the transition diameter as a function of gravity
        # The transition values come from CTEM and are a synthesis of Pike (1980), Croft (1985), Schenk et al. (2004).
        if self.target.transition_scale_type == "silicate":
            simple_complex_exp = -1.0303 
            simple_complex_mean = 2*16533.8 
            simple_complex_std = 0.04
        elif self.target.transition_scale_type == "ice":
            simple_complex_exp = -1.22486
            simple_complex_mean = 2*3081.39
            simple_complex_std = 0.04
        
        # The nominal value will be used for determining the range of the "transitional" morphology type
        transition_nominal= simple_complex_mean * self.target.gravity**simple_complex_exp
        
        # Draw from a truncated normal distribution for each component of the model
        simple_enlargement_factor = 1.0 / mc.bounded_norm(simple_enlargement_mean, simple_enlargement_std)
        final_exp = mc.bounded_norm(final_exp_mean, final_exp_std)
        simple_complex_fac = simple_complex_mean * np.exp(self.rng.normal(loc=0.0,scale=simple_complex_std))
        transition_diameter = simple_complex_fac * self.target.gravity**simple_complex_exp
        self.transition_diameter = transition_diameter
        self.transition_nominal=transition_nominal
        self.simple_enlargement_factor = simple_enlargement_factor
        self.complex_enlargement_factor = complex_enlargement_factor
        self.final_exp = final_exp
        return 


    def get_morphology_type(self, final_diameter: FloatLike) -> str:
        """
        Computes and the morphology type of a crater and returns a string corresponding to its type.

        Parameters
        ----------
        final_diameter : float
            The diameter of the crater to compute
        
        Returns
        ----------
        str
            The type of crater "simple", "complex", or "transitional" 
        """
        
        # Use the 1/2x to 2x the nominal value of the simple->complex transition diameter to get the range of the "transitional" morphology type. This is supported by: Schenk et al. (2004) and Pike (1980) in particular  
        transition_range = (0.5*self.transition_nominal,2*self.transition_nominal)
        
        if final_diameter < transition_range[0]:
            morphology_type = "simple" 
        elif final_diameter > transition_range[1]:
            morphology_type = "complex"
        else:
            # We'll uses the distance from the nominal transition diameter to set a probability of being either simple, complex, or transitional.
            if final_diameter < self.transition_nominal:
                p = (self.transition_nominal - final_diameter)/(self.transition_nominal - transition_range[0])
                categories = ["simple","transitional"]
                prob = [p, 1.0-p] 
                morphology_type = self.rng.choice(categories,p=prob)
            else:
                p = (final_diameter - self.transition_nominal)/(transition_range[1] - self.transition_nominal)
                categories = ["complex","transitional"]
                prob = [p, 1.0-p] 
                morphology_type = self.rng.choice(categories,p=prob)                
        
        return morphology_type


    def f2t_simple(self, Df):
        return Df / self.simple_enlargement_factor
    
    
    def f2t_complex(self, Df):
        return Df / (self.simple_enlargement_factor * self.complex_enlargement_factor) * (Df / self.transition_diameter)**-self.final_exp
    
    
    def final_to_transient(self, final_diameter: FloatLike, morphology_type: str | None = None, **kwargs) -> np.float64:
        """
        Computes the transient diameter of a crater based on its final diameter and morphology type.

        This method first ensures that the morphology type of the crater is computed. It then calculates
        the transient crater diameter based on the final diameter using scaling factors for simple or complex
        crater morphologies.

        Parameters
        ----------
        final_diameter : float-like
            The final crater diameter in meters for which to compute the transient diameter.
        morphology_type : str, optional
            The morphology type of the crater ("simple", "complex", "transitional")

        Returns
        -------
        np.float64
            Returns the crater transient diameter in meters
        """        
        if not morphology_type:
            morphology_type = self.get_morphology_type(final_diameter) 
        
        if morphology_type == "simple": 
            transient_diameter = self.f2t_simple(final_diameter)
        else:
            transient_diameter = self.f2t_complex(final_diameter)

        transient_diameter = np.float64(transient_diameter)
        return transient_diameter, morphology_type


    def transient_to_final(self, transient_diameter: FloatLike) -> Tuple[np.float64, str]:
        """
        Computes the final diameter of a crater based on its transient diameter and morphology type.

        This method first ensures that the morphology type of the crater is computed. It then calculates
        the final crater diameter based on the transient diameter using scaling factors for simple or complex
        crater morphologies. This is a bit more complicated than the final->transient calculation  because In 
        the transition region, a particular transient crater diameter could be associate with simple, complex, 
        or transitional crater morphologies. Therefore we need to monte carlo our way into a solution to avoid 
        biasing in favor of one or another in the transient->final computation

        Parameters
        ----------
        transient_diameter : float-like
            The transient diameter in meters of the crater to convert to final

        Returns
        -------
        np.float64
            The final crater diameter
        str
            The morphology type of the crater
        """ 
        
        # Invert the final -> transient functions for  each crater type
        final_diameter_simple = transient_diameter * self.simple_enlargement_factor
        def root_func(final_diameter,Dt,scale):
            return scale.f2t_complex(final_diameter) - Dt
            
        sol = root_scalar(lambda x, *args: root_func(x, *args),bracket=(0.1*final_diameter_simple,10*final_diameter_simple), args=(transient_diameter, self))
        final_diameter_complex = sol.root
        
        # Evaluate the potential morphology that this transient crater could be consistent with. If both potential diameter values are unambigusously is unambiguosuly simple or complex, go with that.
        # If there is disagreement, then we'll draw the answer from a hat and just check to make sure that final_diameter > transient_diameter 
        morphology_options = [self.get_morphology_type(final_diameter_simple),self.get_morphology_type(final_diameter_complex)]
        
        if len(set(morphology_options)) == 1: # We have agreement!
            morphology_type = morphology_options[0]
            if morphology_type == "simple":
                final_diameter = final_diameter_simple
            else:
                final_diameter = final_diameter_complex # this includes transitional types as well
        else: 
            if "simple" in morphology_options: # The disagreement is between simple/complex or simple/transitional
                if morphology_options[0] == "simple":
                    sind = 0
                    cind = 1 
                else:
                    sind = 1
                    cind = 0
                    
                # Randomly draw a morphology based on weighting by whichever option is closest to the transition 
                is_simple = self.rng.random() < np.abs(final_diameter_complex - self.transition_diameter) / np.abs(final_diameter_simple - final_diameter_complex)
                if is_simple:
                    final_diameter = final_diameter_simple
                    morphology_type = morphology_options[sind] 
                else:
                    final_diameter = final_diameter_complex
                    morphology_type = morphology_options[cind]
            else:
                final_diameter = final_diameter_complex
                morphology_type = self.rng.choice(morphology_options)
        
        final_diameter = np.float64(final_diameter)
        morphology_type = morphology_type
        return final_diameter, morphology_type
 
        
    def projectile_to_crater(self, projectile, **kwargs):
        """
        Convert a projectile to its corresponding crater.

        Parameters
        ----------
        projectile : Projectile
            The projectile to be converted.
        target : Target
            The target body being impacted
        Returns
        -------
        Crater
            The crater resulting from the impact of the projectile.
        """
        from .impact import Crater
        transient_diameter = self.projectile_to_transient(projectile, target=self.target, material=self.material, rng=self.rng)
        crater = Crater(transient_diameter=transient_diameter, target=self.target, rng=self.rng, **kwargs, location=projectile.location, age=projectile.age)

        return crater


    def crater_to_projectile(self, crater, **kwargs):
        """
        Convert a crater back to its corresponding projectile.
        This operation is more hypothetical and approximates the possible projectile that created the crater.

        Parameters
        ----------
        crater : Crater
            The crater to be converted.

        Returns
        -------
        Projectile
            The estimated projectile that could have caused the crater.
        """
        projectile = self.transient_to_projectile(crater, rng=self.rng, **kwargs)
        
        return projectile


    @staticmethod
    def projectile_to_transient(projectile, 
                                target: Target,
                                material: Material,
                                **kwargs: Any) -> np.float64:
        """
        Calculate the transient diameter of a crater based on the properties of the projectile and target.

        Parameters
        ----------
        projectile : Projectile
            The projectile responsible for the impact.
        target : Target
            The target body for the impact simulation.
        material : Material
            The material composition of the target body.
        **kwargs : Any
            Additional keyword arguments that might influence the calculation.

        Returns
        -------
        np.float64
            The calculated transient diameter of the crater resulting from the impact.
        """
        from .impact import Projectile
        if not isinstance(projectile, Projectile):
            raise TypeError("projectile must be an instance of Projectile")
        if not isinstance(target, Target):
            raise TypeError("target must be an instance of Target")
        if not isinstance(material, Material):
            raise TypeError("material must be an instance of Material")
            
        # Compute some auxiliary quantites
        projectile.mass = 4.0/3.0 * np.pi * projectile.density * (projectile.radius)**3
        mu = material.mu
        kv = material.K1
        c1 = 1.0 + 0.5 * mu
        c2 = (-3 * mu)/(2.0 + mu)

        # Find dimensionless quantities
        pitwo = (target.gravity * projectile.radius)/(projectile.vertical_velocity**2)
        pithree = material.Ybar / (material.density * (projectile.vertical_velocity**2))
        pifour = material.density / projectile.density
        pivol = kv * ((pitwo * (pifour**(-1.0/3.0))) + (pithree**c1))**c2
        pivolg = kv * (pitwo * (pifour**(-1.0/3.0)))**c2
        
        # find transient crater volume and radii (depth = 1/3 diameter)
        cvol = pivol * (projectile.mass / material.density)
        cvolg = pivolg * (projectile.mass / material.density)
        transient_radius = (3 * cvol / np.pi)**(1.0/3.0)
        transient_radius_gravscale = (3 * cvolg / np.pi)**(1.0/3.0)
        
        transient_diameter = transient_radius * 2
        
        if transient_diameter < projectile.diameter:
            transient_diameter = projectile.diameter
        
        return transient_diameter


    def transient_to_projectile(self, 
                                crater, 
                                rng: Generator = None, 
                                **kwargs: Any): 
        """
        Estimate the characteristics of the projectile that could have created a given crater.

        This method approximates the properties of a hypothetical projectile based on the characteristics
        of a known crater.

        Parameters
        ----------
        crater : Crater
            The crater for which to estimate the projectile.
        rng : Generator, optional
            Random number generator instance used for any probabilistic calculations.
        **kwargs : Any
            Additional keyword arguments that might influence the calculation.

        Returns
        -------
        Projectile
            The computed projectile for the crater.
        """
        from .impact import Crater, Projectile
        if not isinstance(crater, Crater):
            raise TypeError("crater must be an instance of Crater")
        if rng and not isinstance(rng, Generator):
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
                
        # We'll create a Projectile object that will allow us to set velocity
        # First pop off any pre-computed values so that we use the transient_diameter alone
        kwargs.pop("diameter",None)
        kwargs.pop("location",None)
        kwargs.pop("age",None)
        projectile = Projectile(diameter=crater.transient_diameter, target=self.target, location=crater.location, rng=rng, age=crater.age, **kwargs)
        
        def root_func(projectile_diameter: FloatLike, 
                      projectile: Projectile, 
                      target: Target,
                      material: Material
                      ) -> np.float64:
            
            projectile.diameter = projectile_diameter
            transient_diameter = self.projectile_to_transient(projectile, target, material)
            return transient_diameter - crater.transient_diameter 
        
        sol = root_scalar(lambda x, *args: root_func(x, *args),bracket=(1e-5*crater.transient_diameter,1.2*crater.transient_diameter), args=(projectile, self.target, self.material))
        
        # Regenerate the projectile with the new diameter value
        projectile = Projectile(diameter=sol.root, target=self.target, location=projectile.location, velocity=projectile.velocity, angle=projectile.angle, direction=projectile.direction, rng=rng, age=projectile.age)
        
        return projectile

    @property
    def transition_diameter(self) -> np.float64:
        """
        The transition diameter between simple and complex craters in m.
        
        Returns
        -------
        np.float64
        """
        return self._transition_diameter
    
    @transition_diameter.setter
    def transition_diameter(self, value: FloatLike) -> None:
        self._transition_diameter = np.float64(value)
        
    @property
    def transition_nominal(self) -> np.float64:
        """
        The nominal transition diameter for crater morphology in m.
        
        Returns
        -------
        np.float64
        """
        return self._transition_nominal
    
    @transition_nominal.setter
    def transition_nominal(self, value: FloatLike) -> None:
        self._transition_nominal = np.float64(value)

    @property
    def simple_enlargement_factor(self) -> np.float64:
        """
        The enlargement factor for simple craters.
        
        Returns
        -------
        np.float64
        """
        return self._simple_enlargement_factor
    
    @simple_enlargement_factor.setter
    def simple_enlargement_factor(self, value: FloatLike) -> None:
        self._simple_enlargement_factor = np.float64(value)

    @property
    def complex_enlargement_factor(self) -> np.float64:
        """
        The enlargement factor for complex craters.
        
        Returns
        -------
        np.float64
        """
        return self._complex_enlargement_factor
    
    @complex_enlargement_factor.setter
    def complex_enlargement_factor(self, value: FloatLike) -> None:
        self._complex_enlargement_factor = np.float64(value)

    @property
    def final_exp(self) -> np.float64:
        """
        The exponent used in the final rim radius to simple crater radius relationship.
        
        Returns
        -------
        np.float64
        """
        return self._final_exp
    
    @final_exp.setter
    def final_exp(self, value: FloatLike) -> None:
        self._final_exp = np.float64(value)

    @property
    def target(self):
        """
        The target body for the impact.
        
        Returns
        -------
        Target
        """ 
        return self._target
    
    @target.setter
    def target(self, value):
        if value is None:
            self._target = Target(name="Moon")
            return
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value
        return 
    

    @property
    def material_name(self):
        """
        The name of the material composition of the target body.
        
        Returns
        -------
        str 
        """
        return self._material_name

    @material_name.setter
    def material_name(self, value):
        if not isinstance(value, str) and value is not None:
            raise TypeError("material_name must be a string or None")
        self._material_name = value
        
        
    @property
    def material(self):
        """
        The material properties associated with the target body.
        
        Returns
        -------
        Material
        """
        return self._material
    
    @material.setter
    def material(self, value):
        if not isinstance(value, Material) and value is not None:
            raise TypeError("material must be a Material object or None")
        self._material = value    
    
    @property
    def rng(self):
        """
        A random number generator instance.
        
        Returns
        -------
        Generator
        """ 
        return self._rng
    
    @rng.setter
    def rng(self, value):
        if not isinstance(value, Generator) and value is not None:
            raise TypeError("The 'rng' argument must be a numpy.random.Generator instance or None")
        self._rng = value or np.random.default_rng()       
      



import unittest
from cratermaker import Target
from cratermaker import create_catalogue
import numpy as np
gEarth = np.float64(9.80665)

class TestTarget(unittest.TestCase):

    def test_target_from_catalogue(self):
        # Test creating a target from the built-in catalogue
        planets = ["Mercury", "Venus", "Earth", "Moon", "Mars", "Ceres", "Vesta" ]
        for planet in  planets:
            target = Target(name=planet)
            v = target.catalogue[planet]
            self.assertEqual(target.radius, v['radius'])
            self.assertEqual(target.gravity, v['gravity'])
            self.assertEqual(target.material_name, v['material_name'])
            self.assertEqual(target.transition_scale_type, v['transition_scale_type'])
            self.assertIsInstance(target.radius, np.float64)
            self.assertIsInstance(target.gravity, np.float64)
            self.assertIsInstance(target.material_name, str)
            self.assertIsInstance(target.transition_scale_type, str)
            
        # Test creating a target from a custom catalogue
        body_properties = [
            "name",    "radius",   "gravity",      "material_name", "transition_scale_type"
        ]
        body_values = [
            ("Arrakis", 5995.0e3,  0.900 * gEarth, "Sand", "silicate"),
            ("Salusa Secundus", 7200.e3, 1.20 * gEarth, "Hard Rock", "silicate"),
        ]          
        catalogue = create_catalogue(body_properties, body_values) 
        for planet,v in catalogue.items():
            target = Target(name=planet, catalogue=catalogue)
            self.assertEqual(target.radius, v['radius'])
            self.assertEqual(target.gravity, v['gravity'])
            self.assertEqual(target.material_name, v['material_name'])
            self.assertEqual(target.transition_scale_type, v['transition_scale_type'])
            self.assertIsInstance(target.radius, np.float64)
            self.assertIsInstance(target.gravity, np.float64)


    def test_custom_target(self):
        # Test creating a custom target
        target = Target(name="Arrakis", radius=5995.0e3, gravity=0.9*gEarth, material_name="Sand", transition_scale_type="silicate")
        self.assertEqual(target.radius, 5995.0e3)
        self.assertEqual(target.gravity, 0.9*gEarth)
        self.assertEqual(target.material_name, "Sand")

    def test_invalid_target(self):
        # Test incomplete custom target creation or invalid arguments
        with self.assertRaises(ValueError):
            Target("Arrakis")
        
        # Missing name
        with self.assertRaises(TypeError):
            Target(radius=5995.0, gravity=0.9*gEarth, material_name="Sand", transition_scale_type="silicate")
        with self.assertRaises(ValueError):
            Target("Moon",transition_scale_type="flubber")

if __name__ == '__main__':
    unittest.main()


import numpy as np
from numpy.random import Generator
import xarray as xr
import os
import shutil
from tqdm import tqdm
from glob import glob
from typing import Any, Tuple, Type, Sequence
from numpy.typing import ArrayLike
import warnings
from .target import Target
from .impact import Crater, Projectile
from .surface import Surface, save
from .scale import Scale
from .morphology import Morphology
from .production import Production, NeukumProduction
from ..utils.general_utils import set_properties
from ..utils.custom_types import FloatLike, PairOfFloats
from ..realistic import apply_noise


class Simulation:
    """
    This class orchestrates the processes involved in running a crater simulation.

    """
    def __init__(self, *, # Enforce keyword-only arguments
                 target: str | Target = None,
                 surf: Surface | None = None,
                 reset_surface: bool = True,
                 seed: int | None = None,
                 rng: Generator | None = None,
                 simdir: os.PathLike | None = None, 
                 scale_cls: Type[Scale] | None = None,
                 morphology_cls: Type[Morphology] | None = None,
                 production_cls: Type[Production] | None = None,
                 ejecta_truncation: FloatLike | None = None,
                 dorays: bool = True,
                 **kwargs: Any):
        """
        Initialize the Simulation object.

        Parameters
        ----------
        target: str or Target, optional, default "Moon"
            Name of the target body or Target object for the simulation, default is "Moon".
        surf : Surface, optional
            Surface object to use for the simulation, default is None, which reads in a surface from file if it exists, or creates a new one.
        reset_surface : bool, optional
            Flag to reset the surface elevation, default is True.
        seed : int, optional
            Seed for the random number generator, default is None.
        rng : Generator, optional
            Random number generator, default is None, which will generate a new np.random generator based on the seed.
        simdir: PathLike, optional
            Path to the simulation directory, default is current working directory.
        scale_cls : Type[Scale], optional
            The Scale class that defines the crater scaling law. If none provided, then the default will be 
        morphology_cls : Type[Morphology], optional
            The Morphology class that defines the model used to describe the morphology of the crater. If none provided, then the 
            default will be based on the default morphology model.
        production_cls: Type[Production], optional
            The Production class that defines the production function used to populate the surface with craters. If none provided, 
            then the default will be based on the target body, with the NeukumProduction crater-based scaling law used if the target 
            body is the Moon or Mars, the NeukumProduction projectile-based scaling law if the target body is Mercury, Venus, or 
            Earth, and a simple power law model otherwise.
        ejecta_truncation : float, optional
            The relative distance from the rim of the crater to truncate the ejecta blanket, default is None, which will compute a 
            truncation distance based on where the ejecta thickness reaches a small value. 
        dorays : bool, optional
            Flag to enable or disable the ejecta ray pattern model, default is True.
        **kwargs : Any
            Additional keyword arguments that can be passed to other the methods of the class, such as arguments to set the scale, 
            morphology, or production function constructors. These also include the grid configuration parameters, such as 
            `grid_type`, and its associated parameters. Refer to the documentation of the surface module for details.
            
        See Also
        --------
        cratermaker.core.surface.Surface.initialize : Parameters for initializing a surface mesh.
        """
     
        self.simdir = simdir
        
        # Set the random number generator seed
        self.seed = seed
        self.rng = rng
        
        self._crater = None
        self._projectile = None
        self._interval_number = 0
        self._elapsed_time = 0.0
        self._current_age = 0.0
        self._elapsed_n1 = 0.0
        self._ejecta_truncation = ejecta_truncation
        self._smallest_crater = 0.0 # The smallest crater will be determined by the smallest face area
        self._smallest_projectile = 0.0 # The smallest crater will be determined by the smallest face area
        self._largest_crater = np.inf # The largest crater will be determined by the target body radius
        self._largest_projectile = np.inf # The largest projectile will be determined by the target body radius
        self._dorays = dorays
         
        # First we need to establish the production function. This will allow us to compute the mean impact velocity, which is needed
        # in order to instantiate the target body.
        #  
        # Check to see if the impact velocity model is set with an argument. If not, set it based on a combination of target body 
        # and production function 
        
        if not target:
            target_name = "Moon"
        elif isinstance(target, str):
            target_name = target
        elif isinstance(target, Target):
            target_name = target.name
        else:
            raise TypeError("target must be an instance of Target or a valid name of a target body")
       
        # Process the impact velocity model and set defaults based on the target body 
        impact_velocity_model = kwargs.get('impact_velocity_model', None)
        if impact_velocity_model is None:
            if target_name in ['Ceres', 'Vesta']:
                impact_velocity_model = "MBA_MBA"
            elif target_name in ['Mercury', 'Venus', 'Earth', 'Moon', 'Mars']:
                impact_velocity_model = target_name + "_MBA"
            kwargs['impact_velocity_model'] = impact_velocity_model
                
        # Set the production function model for this simulation 
        if production_cls is None:
            if target_name in ['Mercury', 'Venus', 'Earth', 'Moon', 'Mars']:
                production_cls = NeukumProduction
                if target_name in ['Moon', 'Mars']:
                    self.production = production_cls(model=target_name, rng=self.rng, **kwargs) 
                else:
                    self.production = production_cls(model='projectile', rng=self.rng, **kwargs)
            else:
                self.production = Production(rng=self.rng, **kwargs)
        elif issubclass(production_cls, Production):
            self.production = production_cls(rng=self.rng, **kwargs)
        else:
            raise TypeError("production must be a subclass of Production")
            
        if not target:
            target = Target("Moon",**kwargs)
        elif isinstance(target, str):
            try:
                target = Target(target,**kwargs)
            except:
                raise ValueError(f"Invalid target name {target}")
        elif not isinstance(target, Target):
            raise TypeError("target must be an instance of Target or a valid name of a target body")
        self.target = target

        # Set the scaling law model for this simulation 
        if scale_cls is None:
            self.scale_cls = Scale
        elif issubclass(scale_cls, Scale):
            self.scale_cls = scale_cls
        else:
            raise TypeError("scale must be a subclass of Scale") 
      
        # Set the morphology model for this simulation 
        if morphology_cls is None:
            self.morphology_cls = Morphology 
        elif issubclass(morphology_cls, Morphology):
            self.morphology_cls = morphology_cls
        else:
            raise TypeError("morphology must be a subclass of Morphology")
      
        self._craterlist = []
        
        if not surf:
            grid_type = kwargs.get('grid_type', None)
            if grid_type is not None and grid_type == 'hires local':
                if 'superdomain_scale_factor' not in kwargs:
                    # Determine the scale factor for the superdomain based on the smallest crater whose ejecta can reach the edge of the 
                    # superdomain. This will be used to set the superdomain scale factor. TODO: Streamline this a bit
                    for d in np.logspace(np.log10(self.target.radius*2), np.log10(self.target.radius / 1e6), 1000):
                        crater, _ = self.generate_crater(diameter=d, angle=90.0, velocity=self.production.mean_velocity*10)
                        rmax = crater.morphology.compute_rmax(minimum_thickness=1e-3) 
                        if rmax < self.target.radius * 2 * np.pi:
                            superdomain_scale_factor = rmax / crater.radius
                            break
                    kwargs['superdomain_scale_factor'] = superdomain_scale_factor
            self.surf = Surface.initialize(target=self.target, reset_surface=reset_surface, simdir=simdir, rng=self.rng, **kwargs)
        elif isinstance(surf, Surface):
            self.surf = surf
        else:
            raise TypeError("surf must be an instance of Surface or None")        
       
        return

    
    def set_properties(self, 
                       **kwargs: Any
                      ) -> None:
        """
        Set properties of the current object based on the provided keyword arguments.

        This function is a utility to update the properties of the current object. The actual implementation of the 
        property setting is handled by the `utils.set_properties` method.

        Parameters
        ----------
        **kwargs : dict
            A dictionary of keyword arguments that represent the properties to be set on the current object.

        """        
        set_properties(self,**kwargs)
        return 

    
    def to_json(self, 
                filename: os.PathLike,
                ) -> str:
        """
        Export the current simulation configuration to a JSON file.

        Parameters
        ----------
        filename : str
            The file path where the JSON configuration will be saved.
        """        
        #TODO: Re-do this once the dust settles a bit
        # Get the simulation configuration into the correct structure
        # target_config = {**to_config(self.target)}
        # sim_config = {**to_config(self),'target' : target_config} 
        
        # # Write the combined configuration to a JSON file
        # with open(filename, 'w') as f:
        #     json.dump(sim_config, f, indent=4)
            
        # return
        pass
   
    def get_smallest_diameter(self, 
                              face_areas: ArrayLike | None = None, 
                              from_projectile: bool = False) -> float:
        """
        Get the smallest possible crater or projectile be formed on the surface.
        """
        if face_areas is None:
            face_areas = self.surf['face_areas'].values
        else:
            face_areas = np.asarray(face_areas)
        smallest_crater = np.sqrt(face_areas.min().item() / np.pi) * 2        
        if from_projectile:
            _, projectile = self.generate_crater(diameter=smallest_crater, angle=90.0, velocity=self.production.mean_velocity*10)
            return projectile.diameter 
        else:
            return smallest_crater 
        
    
    def get_largest_diameter(self,
                             from_projectile: bool = False) -> float:
        """
        Get the largest possible crater or projectile that can be formed on the surface.
        """
        largest_crater = self.target.radius * 2
        if from_projectile:
            _, projectile = self.generate_crater(diameter=largest_crater, angle=1.0, velocity=self.production.mean_velocity/10.0)
            return projectile.diameter
        else:
            return largest_crater

    def generate_crater(self, 
                        **kwargs: Any
                       ) -> Tuple[Crater, Projectile]:
        """
        Create a new Crater object and its corresponding Projectile.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments for initializing the :class:`Crater` object. Refer to 
            its documentation for details on valid keyword arguments.

        Returns
        -------
        tuple of (Crater, Projectile)
            A tuple containing the newly created Crater and Projectile objects.
            
        Notes
        -----
        The keyword arguments provided are passed to the constructor of 
        :class:`Crater`. Additionally, these arguments are used in 
        :meth:`crater.scale.crater_to_projectile` method. Refer to the 
        documentation of these classes and methods for a detailed description 
        of valid keyword arguments.

        Examples
        --------
        .. code-block:: python
        
            # Create a crater and projectile pair with a specific diameter
            crater, projectile = sim.generate_crater(diameter=1000.0)
    
            # Create a crater with a specific transient diameter and location (but ignore the projectile)
            crater, _ = sim.generate_crater(transient_diameter=5e3, location=(43.43, -86.92))
        """       
        # Create a new Crater object with the passed arguments and set it as the crater of this simulation
        crater = Crater(target=self.target, morphology_cls=self.morphology_cls, scale_cls=self.scale_cls, rng=self.rng, dorays=self.dorays, **kwargs)
        
        if "velocity" not in kwargs and "mean_velocity" not in kwargs and "vertical_velocity" not in kwargs:
            if self.production.mean_velocity is None:
                projectile = None
                return crater, projectile
            kwargs['mean_velocity'] = self.production.mean_velocity
        projectile = crater.scale.crater_to_projectile(crater,**kwargs)
        
        return crater, projectile
    
    
    def generate_projectile(self, 
                            **kwargs: Any
                           ) -> Tuple[Projectile, Crater]:
        """
        Create a new Projectile object and its corresponding Crater.

        Parameters
        ----------
        **kwargs : Any
            Keyword arguments for initializing the :class:`Projectile` object. Refer to its documentation 
            for details on valid keyword arguments.

        Returns
        -------
        (Projectile, Crater)
            A tuple containing the newly created Projectile and Crater objects.
            
        Notes
        -----
        The keyword arguments provided are passed to the constructor of 
        :class:`Projectile`. Refer to its documentation for a detailed description 
        of valid keyword arguments.    
        
        Examples
        --------
        .. code-block:: python
        
            # Create a projectile and crater pair with a specific diameter
            projectile, crater = sim.generate_projectile(diameter=1000.0)

            # Create a projectile and crater with a specific mass and velocity and impact angle
            projectile, crater = sim.generate_projectile(mass=1e14, velocity=20e3, angle=10.0)       
        """
        if "velocity" not in kwargs and "mean_velocity" not in kwargs and "vertical_velocity" not in kwargs:
            if self.production.mean_velocity is None:
                raise RuntimeError("No velocity value is set for this projectile")
            kwargs['mean_velocity'] = self.production.mean_velocity        
        
        projectile = Projectile(target=self.target, rng=self.rng, scale_cls=self.scale_cls,**kwargs)
        crater = projectile.scale.projectile_to_crater(projectile, morphology_cls=self.morphology_cls)
        
        return projectile, crater
   
   
    def emplace_crater(self, 
                       from_projectile: bool=False, 
                       **kwargs: Any
                      ) -> None:
        """
        Emplace a crater in the simulation, optionally based on a projectile.

        This method orchestrates the creation and placement of a crater in the
        simulation. It can create a crater directly or based on the characteristics
        of a projectile.

        Parameters
        ----------
        from_projectile : bool, optional
            Flag to create a crater based on a projectile, default is False.
        **kwargs : Any
            Keyword arguments for initializing the :class:`Crater` or 
            :class:`Projectile` object. Refer to the documentation of these 
            classes for details on valid keyword arguments.

        Notes
        -----
        The keyword arguments provided are passed down to :meth:`generate_crater` 
        or :meth:`generate_projectile`, and subsequently to the constructors of 
        :class:`Crater`, :class:`Projectile`, or :class:`Impact`. Refer to the 
        documentation of these classes for a detailed description of valid 
        keyword arguments.

        Examples
        --------
        .. code-block:: python        
        
            # Create a crater with specific diameter
            sim.emplace_crater(diameter=1000.0)

            # Create a crater based on a projectile with given mass and velocity
            sim.emplace_crater(from_projectile=True, mass=1e14, velocity=20e3)
            
            # Create a crater with a specific transient diameter and location
            sim.emplace_crater(transient_diameter=5e3, location=(43.43, -86.92))
        """ 
        if from_projectile:
            self.projectile, self.crater = self.generate_projectile(**kwargs)
        else:
            self.crater, self.projectile = self.generate_crater(**kwargs)
       
        self.crater.node_index, self.crater.face_index = self.surf.find_nearest_index(self.crater.location)
        self.projectile.node_index, self.projectile.face_index = self.crater.node_index, self.crater.face_index
        self.crater.morphology.form_crater(self.surf,**kwargs)
        self.crater.morphology.form_ejecta(self.surf,**kwargs)
        if self.dorays:
            self.crater.morphology.form_secondaries(self.surf,**kwargs)
        
        return  


    def populate(self, 
                 age: FloatLike | None = None,
                 age_end: FloatLike | None = None,
                 diameter_number: PairOfFloats | None = None,
                 diameter_number_end: PairOfFloats | None = None,
                 **kwargs: Any,
                ) -> None:
        """
        Populate the surface with craters over a specified interval using the current production function.
        
        Parameters
        ----------
        age : FloatLike, optional
            Age in the past in units of My relative to the present, which is used compute the cumulative SFD.
        age_end, FloatLike, optional
            The ending age in units of My relative to the present, which is used to compute the cumulative SFD. The default is 0 (present day).
        diameter_number : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N). If provided, the function will convert this value
            to a corresponding age and use the production function for a given age.
        diameter_number_end : PairOfFloats, optional
            A pair of diameter and cumulative number values, in the form of a (D, N). If provided, the function will convert this
            value to a corresponding reference age and use the production function for a given age.        
        """

        if not hasattr(self, 'production'):
            raise RuntimeError("No production function defined for this simulation")
        elif not hasattr(self.production, 'generator_type'):
            raise RuntimeError("The production function is not properly defined. Missing 'generator_type' attribute")
        elif self.production.generator_type not in ['crater', 'projectile']:
            raise RuntimeError(f"Invalid production function type {self.production.generator_type}")
       
        from_projectile = self.production.generator_type == 'projectile'
        Dmax = self.get_largest_diameter(from_projectile=from_projectile)
        Dmin = self.get_smallest_diameter(from_projectile=from_projectile)
        
        # Loop over each face in the mesh to build up a population of craters in this interval. This is done because faces may
        # not all have the same surface area, the range of crater sizes that can be formed on each face may be different.
        impact_diameters = []
        impact_ages = []
        impact_locations = []
        face_areas = self.surf['face_areas'].values
        min_area = face_areas.min()
        n_face = face_areas.size
        surface_area = self.surf.area.item() 
         
        # Group surfaces into bins based on their area. All bins within a factor of 2 in surface area are grouped together.
        max_bin_index = np.ceil(np.log2(face_areas.max() / min_area)).astype(int)
        bins = {i: [] for i in range(max_bin_index + 1)}
        
        for face_index, area in enumerate(face_areas):
            bin_index = np.floor(np.log2(area / min_area)).astype(int)
            bins[bin_index].append(face_index)        
            
        # Process each bin
        for bin_index, face_indices in bins.items():
            if not face_indices:
                continue  # Skip empty bins
            bin_areas = face_areas[face_indices]
            total_bin_area = bin_areas.sum()
            area_ratio = total_bin_area / surface_area 
             
            Dmin = self.get_smallest_diameter(bin_areas, from_projectile=from_projectile)
            if diameter_number is not None:
                diameter_number_local = (diameter_number[0], diameter_number[1] * area_ratio)
            else:
                diameter_number_local = None
            if diameter_number_end is not None:
                diameter_number_end_local = (diameter_number_end[0], diameter_number_end[1] * area_ratio)
            else:
                diameter_number_end_local = None
        
            diameters, ages= self.production.sample(age=age, 
                                                    age_end=age_end, 
                                                    diameter_number=diameter_number_local, 
                                                    diameter_number_end=diameter_number_end_local, 
                                                    diameter_range=(Dmin, Dmax),
                                                    area=total_bin_area,
                                                    **kwargs)
            if diameters.size > 0:
                impact_diameters.extend(diameters.tolist())
                impact_ages.extend(ages.tolist())
                
                # Get the probability of impact onto any particular face then get the locations of the impacts
                p = bin_areas / total_bin_area
                locations = []
                for i, d in enumerate(diameters): 
                    face_index = self.rng.choice(face_indices, p=p)
                    locations = self.surf.get_random_location_on_face(face_index)
                    impact_locations.append(locations) 
            
        if len(impact_diameters) > 0: 
            if len(impact_diameters) == 1:
                self.emplace_crater(diameter=impact_diameters[0], age=impact_ages[0], location=impact_locations[0], from_projectile=from_projectile)
            else:
                # Sort the ages, diameters, and locations so that they are in order of decreasing age
                sort_indices = np.argsort(impact_ages)[::-1]
                impact_diameters = np.asarray(impact_diameters)[sort_indices]
                impact_ages = np.asarray(impact_ages)[sort_indices] 
                impact_locations = np.array(impact_locations, dtype=[('lon', 'float64'), ('lat', 'float64')])[sort_indices]
                
                for i, diameter in tqdm(enumerate(impact_diameters), total=len(impact_diameters)):
                    location = impact_locations[i][0], impact_locations[i][1]
                    self.emplace_crater(diameter=diameter, age=impact_ages[i], location=location, from_projectile=from_projectile)
        return 
    
    
    def run(self,
            age: FloatLike | None = None,
            age_end: FloatLike | None = None,
            age_interval: FloatLike | None = None,
            diameter_number: PairOfFloats | None = None,
            diameter_number_end: PairOfFloats | None = None,
            diameter_number_interval: FloatLike | None = None,
            ninterval: int | None = None,
            **kwargs: Any,
           ) -> None:
        """
        Run the simulation over a specified interval using the current production function.

        Parameters
        ----------
        age : FloatLike, optional
            Start age in My relative to the present for the simulation, used to compute the starting point of the production function.
            Default is None, which requires `diameter_number` to be set.
        age_end : FloatLike, optional
            End age in My relative to the present for the simulation, used to compute the ending point of the production function.
            Default is 0 (present day) if not provided.
        age_interval : FloatLike, optional
            Interval in My for outputting intermediate results. If not provided, calculated as `age - age_end` / `ninterval` if `ninterval` is provided, otherwise set to the total simulation duration.
        diameter_number : PairOfFloats, optional
            Starting cumulative number and diameter pair (D, N) to define the simulation start in terms of crater frequency. 
            Default is None, which requires `age` to be set.
        diameter_number_end : PairOfFloats, optional
            Ending cumulative number and diameter pair (D, N) to define the simulation end in terms of crater frequency.
            Default is the present-day values if not provided.
        diameter_number_interval : PairOfFloats, optional
            Interval for outputting results in terms of cumulative number and diameter (D, N). Calculated based on `ninterval` if provided.
        ninterval : int, optional
            Number of intervals for outputting results. This has a special use case where one can specify age-based inputs but output
            in equal cumulative number intervals and vice versa.  

        Notes
        -----
        This function allows defining the simulation parameters either in terms of age or crater frequency (cumulative number). The 
        arguments `age`, `age_end`, `age_interval` are mutually exclusive with `diameter_number`, `diameter_number_end`, `diameter_number_interval`.
        
        The input diameter values used in diameter_number, diameter_number_end, and diameter_number_interval need not be the same.
        However, the production function will be used to convert all numbers to a common diameter value on output.

        Examples
        --------
        .. code-block:: python
        
            # Create a simulation object with default parameters (Moon, NeukumProduction, etc.)
            sim = cratermaker.Simulation()
            
            # Run the simulation for 3.8 billion years, saving the results every 100 million years
            sim.run(age=3.8e3, age_interval=100.0)
        
            # Run the simulation to create 80 craters larger than 300 km in diameter and output the results after approximately 100 
            # craters larger than 50km in diameter have formed
            sim.run(diameter_number=(300e3, 80), diameter_number_interval=(50e3, 100))
        
            # Run the simulation for 3.8 billion years, saving 100 intervals with equal cumulative number intervals
            sim.run(age=3.8e3, ninterval=100)
        
            # Run the simulation to create 80 craters larger than 300 km and output 100 intervals with equal age intervals
            sim.run(diameter_number=(300e3, 80), ninterval=100)
        
            # Run the simulation from 3.8 billion years to 3.0 billion years, saving the results every 100 million years
            sim.run(age=3.8e3, age_end=3.0e3, age_interval=100.0)

        """
        arguments = locals().copy()
        arguments.pop('self')
        arguments = self._validate_run_args(**arguments) 
        age = arguments.pop('age', None)
        age_end = arguments.pop('age_end', None)
        age_interval = arguments.pop('age_interval', None)
        diameter_number = arguments.pop('diameter_number', None)
        diameter_number_end = arguments.pop('diameter_number_end', None)
        diameter_number_interval = arguments.pop('diameter_number_interval', None)
        ninterval = arguments.pop('ninterval', None)
        is_age_interval = arguments.pop('is_age_interval', None)
        
        
        self.current_age = age
        self.elapsed_time = 0.0
        self.elapsed_n1 = 0.0
        for i in tqdm(range(ninterval+1), total=ninterval+1):
            self.interval_number = i
            if i > 0: # This allows us to save the initial state of the simulation
                if is_age_interval:
                    current_age = age - (i-1) * age_interval
                    current_age_end = age - i * age_interval
                    self.populate(age=current_age, age_end=current_age_end)
                else:
                    current_diameter_number = (diameter_number[0], diameter_number[1] - (i-1) * diameter_number_interval[1])
                    current_diameter_number_end = (diameter_number[0], diameter_number[1] - i * diameter_number_interval[1])
                    self.populate(diameter_number=current_diameter_number, diameter_number_end=current_diameter_number_end)
                    current_age = self.production.function_inverse(*current_diameter_number)
                    current_age_end = self.production.function_inverse(*current_diameter_number_end)
                    age_interval = current_age - current_age_end 
                self.elapsed_time += age_interval
                self.elapsed_n1 += (self.production.function(diameter=1000.0, age=current_age)  - self.production.function(diameter=1000.0, age = current_age_end))
                self.current_age = current_age_end
                
            self.save()
            
        self.export_vtk()
        return


    def _validate_run_args(self,**kwargs) -> dict:
        """
        Validate all the input arguments to the sample method. This function will raise a ValueError if any of the arguments are invalid.
        It will also convert age arguments to diameter_number and vice versa.
        
        Parameters
        ----------
        age : FloatLike, optional
            Start age in My relative to the present for the simulation, used to compute the starting point of the production function.
            Default is None, which requires `diameter_number` to be set.
        age_end : FloatLike, optional
            End age in My relative to the present for the simulation, used to compute the ending point of the production function.
            Default is 0 (present day) if not provided.
        age_interval : FloatLike, optional
            Interval in My for outputting intermediate results. If not provided, calculated as `age - age_end` / `ninterval` if `ninterval` is provided, otherwise set to the total simulation duration.
        diameter_number : PairOfFloats, optional
            Starting cumulative number and diameter pair (D, N) to define the simulation start in terms of crater frequency. 
            Default is None, which requires `age` to be set.
        diameter_number_end : PairOfFloats, optional
            Ending cumulative number and diameter pair (D, N) to define the simulation end in terms of crater frequency.
            Default is the present-day values if not provided.
        diameter_number_interval : PairOfFloats, optional
            Interval for outputting results in terms of cumulative number and diameter (D, N). Calculated based on `ninterval` if provided.
        ninterval : int, optional
            Number of intervals for outputting results. This has a special use case where one can specify age-based inputs but output
            in equal cumulative number intervals and vice versa.  
                
        Returns
        -------
        A dict containing all arguments listed in Parameters above, as well as `is_age_interval`, which is a boolean flag indicating
        whether or not the simulation is being run in equal age intervals or equal number intervals.
            
        Raises
        ------
        ValueError
            If any of the following conditions are met:
            - Neither the age nore the diameter_number argument is provided.
            - Both the age and diameter_number arguments are provided.
            - Both the age_end and diameter_number_end arguments are provided.
            - The age argument is provided but is not a scalar.
            - The age_end argument is provided but is not a scalar. 
            - The age_interval is provided but is not a positive scalar
            - The age_interval provided is negative, or is greater than age - age_end
            - The diameter_number argument is not a pair of values, or any of them are less than 0
            - The diameter_number_end argument is not a pair of values, or any of them are less than 0
            - The diameter_number_interval argument is not a pair of values, or any of them are less than 0
            - The age_interval and diameter_number_interval arguments are both provided.
            - The diameter_number_interval provided is negative, or is greater than diameter_number - diameter_number_end
            - The ninterval is provided but is not an integer or is less than 1.
            - The ninterval is provided and either age_interval or diameter_number_interval is also provided
        """         
        # Determine whether we are going to do equal time intervals or equal number intervals
        age = kwargs.get('age', None)
        age_interval = kwargs.get('age_interval', None)
        diameter_number_interval = kwargs.get('diameter_number_interval', None)
        ninterval = kwargs.pop('ninterval', None)
        if age_interval is not None and diameter_number_interval is not None:
            raise ValueError("Cannot specify both ninterval and age_interval or diameter_number_interval")
        if ninterval is not None:
            if not isinstance(ninterval, int):
                raise TypeError("ninterval must be an integer")
            if ninterval < 1:
                raise ValueError("ninterval must be greater than zero") 
            if age_interval is not None or diameter_number_interval is not None:
                raise ValueError("Cannot specify both ninterval and age_interval or diameter_number_interval")
            
        is_age_interval = age_interval is not None or (ninterval is not None and diameter_number_interval is None)
        
        # Validate arguments using the production function validator first 
        if 'diameter_range' not in kwargs:
            kwargs['diameter_range'] = (self.get_smallest_diameter(), self.get_largest_diameter())
        if 'area' not in kwargs:
            kwargs['area'] = self.surf.area.item()
        kwargs  = self.production._validate_sample_args(**kwargs)
       
        if is_age_interval:
            age = kwargs.get('age', None)
            if age is None:
                raise ValueError("Something went wrong! age should be set by self.production_validate_sample_args")
            age_end = kwargs.get('age_end', None)
            if age_end is None:
                raise ValueError("Something went wrong! age_end should be set by self.production_validate_sample_args")
            if age_interval is None:
                if ninterval is None:
                    ninterval = 1
                age_interval = (age - kwargs['age_end']) / ninterval
            else:
                if age_interval > age - age_end:
                    raise ValueError("age_interval must be less than age - age_end")
                elif age_interval <= 0:
                    raise ValueError("age_interval must be greater than zero")
                ninterval = int(np.ceil((age - age_end) / age_interval)) 
            
            kwargs['age_interval'] = age_interval
            kwargs['ninterval'] = ninterval
        else:
            diameter_number = kwargs.get('diameter_number', None)
            if diameter_number is None:
                raise ValueError("Something went wrong! diameter_number should be set by self.production_validate_sample_args")
            diameter_number_end = kwargs.get('diameter_number_end', None)
            if diameter_number_end is None:
                raise ValueError("Something went wrong! diameter_number_end should be set by self.production_validate_sample_args")
            if diameter_number_interval is None:
                if ninterval is None:
                    ninterval = 1
                diameter_number_interval = (diameter_number[0], (diameter_number[1] - diameter_number_end[1]) / ninterval)
            else:
                if len(diameter_number_interval) != 2:
                    raise ValueError("The 'diameter_number_interval' must be a pair of values in the form (D,N)")
                # Check to be sure that the diameter in the diameter_number_interval is the same as diameter_number. 
                # If not, we need to adjust the end diameter value it so that they match 
                diameter_number_interval = self.production._validate_csfd(*diameter_number_interval)
                if diameter_number_interval[0] != diameter_number[0]:
                    area = kwargs.get('area', None)
                    if area is None:
                        raise ValueError("Something went wrong! area should be set by self.production_validate_sample_args")
                    diameter_number_density_interval = (diameter_number_interval[0], diameter_number_interval[1] / area)
                    age_val = self.production.function_inverse(*diameter_number_density_interval)
                    diameter_number_density_interval = (diameter_number[0], self.production.function(diameter=diameter_number[0], age=age_val))
                    diameter_number_interval = (diameter_number[0], diameter_number_density_interval[1] * area)
                    
                if diameter_number_interval[1] >= diameter_number[1] - diameter_number_end[1]:
                    raise ValueError("diameter_number_interval must be less than diameter_number - diameter_number_end")
                if diameter_number_interval[1] <= 0:
                    raise ValueError("diameter_number_interval must be greater than zero")
                ninterval = int(np.ceil((diameter_number[1] - diameter_number_end[1]) / diameter_number_interval[1]))
                
            kwargs['diameter_number_interval'] = diameter_number_interval
            kwargs['ninterval'] = ninterval
                
        kwargs['is_age_interval'] = is_age_interval
        
        # Remove unecessary arguments that came out of the production._validate_sample_args method        
        kwargs.pop('diameter_range')
        kwargs.pop('area')
        kwargs.pop('return_age')
         
        return kwargs
   
    
    def save(self, **kwargs: Any) -> None:
        """
        Save the current simulation state to a file.
        """

        time_variables = {
            "current_age": self.current_age,
            "elapsed_time": self.elapsed_time,
            "elapsed_n1": self.elapsed_n1
            }
         
        save(self.surf, interval_number=self.interval_number, time_variables=time_variables, **kwargs)
        
        return
    
    
    def export_vtk(self, 
                   out_dir: os.PathLike | None = None,
                   *args, **kwargs
                   ) -> None:
        """
        Export the surface mesh to a VTK file.

        Parameters
        ----------
        out_dir : str, Default "vtk_files" in the simulation directory
            Directory to store the VTK files.
        """
        from vtk import vtkUnstructuredGrid, vtkPoints, VTK_POLYGON, vtkWarpScalar, vtkXMLPolyDataWriter
        from vtkmodules.util.numpy_support import numpy_to_vtk
        from vtkmodules.vtkFiltersCore import vtkPolyDataNormals
        from vtkmodules.vtkFiltersGeometry import vtkGeometryFilter
        
        self.save()  
        if out_dir is None:
            out_dir = os.path.join(self.simdir, "vtk_files")
        if os.path.exists(out_dir):
            shutil.rmtree(out_dir)
        os.makedirs(out_dir)
            
        data_file_list = glob(os.path.join(self.surf.data_dir, "*.nc"))
        if self.surf.grid_file in data_file_list:
            data_file_list.remove(self.surf.grid_file)
       
        # Convert uxarray grid arrays to regular numpy arrays for vtk processing 
        n_node = self.surf.uxgrid.n_node
        n_face = self.surf.uxgrid.n_face
        node_x = self.surf.uxgrid.node_x.values * self.target.radius
        node_y = self.surf.uxgrid.node_y.values * self.target.radius
        node_z = self.surf.uxgrid.node_z.values * self.target.radius
        n_nodes_per_face = self.surf.uxgrid.n_nodes_per_face.values
        face_node_connectivity = self.surf.uxgrid.face_node_connectivity.values
        
        vtk_data = vtkUnstructuredGrid()
        nodes = vtkPoints()
        for i in range(n_node):
            nodes.InsertNextPoint(node_x[i], node_y[i], node_z[i])
        vtk_data.SetPoints(nodes)
        vtk_data.Allocate(n_face)
        for i,n in enumerate(n_nodes_per_face):
            point_ids=face_node_connectivity[i][0:n]
            vtk_data.InsertNextCell(VTK_POLYGON, n, point_ids) 
       
        warp = vtkWarpScalar()
        warp.SetInputArrayToProcess(0, 0, 0,
                            vtkUnstructuredGrid.FIELD_ASSOCIATION_POINTS,
                            "node_elevation")                
            
        writer = vtkXMLPolyDataWriter()
        writer.SetDataModeToBinary()
        writer.SetCompressorTypeToZLib()
        print("Exporting VTK files...")
        
        with xr.open_mfdataset(data_file_list) as ds:
            # Warp the surface based on node_elevation data
            for i in tqdm(range(len(ds.time))):
                
                ids = ds.isel(time=i)
                current_grid = vtkUnstructuredGrid()
                current_grid.DeepCopy(vtk_data) 
                 
                for v in ds.variables:
                    array = numpy_to_vtk(ids[v].values, deep=True)
                    array.SetName(v)
                    n = ids[v].size
                    if 'n_face' in ids[v].dims:
                        current_grid.GetCellData().AddArray(array)
                    elif 'n_node' in ids[v].dims:
                        current_grid.GetPointData().AddArray(array)
                        if v == 'node_elevation':
                            current_grid.GetPointData().SetActiveScalars(v) 
                    elif n == 1:
                        current_grid.GetFieldData().AddArray(array)
                        
                geomFilter = vtkGeometryFilter()
                geomFilter.SetInputData(current_grid)
                geomFilter.Update()    
                polyData = geomFilter.GetOutput()    

                normalsFilter = vtkPolyDataNormals()
                normalsFilter.SetInputData(polyData)
                normalsFilter.ComputeCellNormalsOn()
                normalsFilter.ConsistencyOn()           # Tries to make normals consistent across shared edges
                normalsFilter.AutoOrientNormalsOn()     # Attempt to orient normals consistently outward/inward
                normalsFilter.SplittingOff()   
                normalsFilter.Update()
                polyDataWithNormals = normalsFilter.GetOutput()        
        
                warp.SetInputData(polyDataWithNormals)
                warp.Update()
                warped_output = warp.GetOutput()
                output_filename = os.path.join(out_dir, f"surf{i:06d}.vtp")
                writer.SetFileName(output_filename)
                writer.SetInputData(warped_output) 
                writer.Write()               
        
        return
    
    def make_circle_file(self,
                         diameters: FloatLike | Sequence[FloatLike] | ArrayLike ,
                         longitudes: FloatLike | ArrayLike,
                         latitudes: FloatLike | ArrayLike,
                         output_filename: os.PathLike | None = None,
                         *args, **kwargs
                        ) -> None:
        """
        Plot circles of diameter D centered at the given location.
    
        Parameters
        ----------
        diameters : FloatLike or ArrayLike
            Diameters of the circles in m.
        longitudes : FloatLike or ArrayLike of Floats
            Longitudes of the circle centers in degrees.
        latitudes : FloatLike or ArrayLike of Floats
            Latitudes of the circle centers in degrees.
        out_dir : str, Default "vtk_files" in the simulation directory
            Directory to store the VTK files.
        """ 
       
        if output_filename is None:
            output_filename = os.path.join(self.simdir, "circles.vtp") 
        else:
            output_filename = os.path.join(self.simdir, output_filename)
        
        diameters = np.atleast_1d(diameters)
        longitudes = np.atleast_1d(longitudes)
        latitudes = np.atleast_1d(latitudes)
        # Check for length consistency
        if len(diameters) != len(longitudes) or len(diameters) != len(latitudes):
                raise ValueError("The diameters, latitudes, and longitudes, arguments must have the same length")
            
        # Validate non-negative values
        if np.any(diameters < 0):
            raise ValueError("All values in 'diameters' must be non-negative")
        
        sphere_radius = self.target.radius 
        def create_circle(lon, lat, circle_radius, num_points=360):
            """
            Create a circle on the sphere's surface with a given radius and center.
            
            Parameters
            ----------
            lon : float
                Longitude of the circle's center in degrees.
            lat : float
                Latitude of the circle's center in degrees.
            circle_radius : float
                Radius of the circle in meters.
            num_points : int, optional
                Number of points to use to approximate the circle. The default is 360. 
            """
            # Create an array of angle steps for the circle
            radians = np.linspace(0, 2 * np.pi, num_points)
            # Convert latitude and longitude to radians
            lat_rad = np.deg2rad(lat)
            lon_rad = np.deg2rad(lon)

            # Calculate the Cartesian coordinates for the circle's center
            center_x = sphere_radius * np.cos(lat_rad) * np.cos(lon_rad)
            center_y = sphere_radius * np.cos(lat_rad) * np.sin(lon_rad)
            center_z = sphere_radius * np.sin(lat_rad)

            # Calculate the vectors for the local east and north directions on the sphere's surface
            east = np.array([-np.sin(lon_rad), np.cos(lon_rad), 0])
            north = np.array([-np.cos(lon_rad)*np.sin(lat_rad), -np.sin(lon_rad)*np.sin(lat_rad), np.cos(lat_rad)])
            
            # Initialize arrays to hold the circle points
            x = np.zeros_like(radians)
            y = np.zeros_like(radians)
            z = np.zeros_like(radians)

            # Calculate the points around the circle
            for i in range(num_points):
                x[i] = center_x + circle_radius * np.cos(radians[i]) * east[0] + circle_radius * np.sin(radians[i]) * north[0]
                y[i] = center_y + circle_radius * np.cos(radians[i]) * east[1] + circle_radius * np.sin(radians[i]) * north[1]
                z[i] = center_z + circle_radius * np.cos(radians[i]) * east[2] + circle_radius * np.sin(radians[i]) * north[2]

            return x, y, z 

        points = vtk.vtkPoints()
        lines = vtk.vtkCellArray()
        point_id = 0  # Keep track of the point ID across all circles
        
        for lon, lat, diameter in zip(longitudes, latitudes, diameters):
            circle_radius = diameter / 2
            x, y, z = create_circle(lon, lat, circle_radius)
            
            for i in range(len(x)):
                points.InsertNextPoint(x[i], y[i], z[i])
            
            polyline = vtk.vtkPolyLine()
            polyline.GetPointIds().SetNumberOfIds(len(x))
            for i in range(len(x)):
                polyline.GetPointIds().SetId(i, point_id)
                point_id += 1
            
            lines.InsertNextCell(polyline)
        
        # Create a polydata object and add points and lines to it
        polydata = vtk.vtkPolyData()
        polydata.SetPoints(points)
        polydata.SetLines(lines)
        
        # Write the polydata to a VTK file
        writer = vtk.vtkXMLPolyDataWriter()
        writer.SetFileName(output_filename)
        writer.SetInputData(polydata)
        
        # Optional: set the data mode to binary to save disk space
        writer.SetDataModeToBinary()
        writer.Write()
        
        return    
    

    def apply_noise(self, 
                    model="turbulence",
                    noise_width=1000e3,
                    noise_height=1e3,
                    to_nodes=True,
                    to_faces=True,
                    **kwargs,
                    ) -> None:
        """
        Applies a specified noise model to the simulation's surface elevation.

        This method adjusts the surface elevation of the simulation based on the chosen noise model. It supports various noise models, each with its own set of default and customizable parameters. The method also ensures that the applied noise is volume-conserving.

        Parameters
        ----------
        model : str, optional
            The noise model to apply. Supported models include 'turbulence', 'billowed', 'plaw', 'ridged', 'swiss', and 'jordan'. The default is 'turbulence'.
        noise_width : float, optional
            The width scale of the noise in meters. The default is 1000e3 (1000 km).
        noise_height : float, optional
            The height scale of the noise in meters. The default is 20e3 (20 km).
        to_nodes : bool, optional
            Flag to apply the noise to the nodes of the mesh. The default is True.
        to_faces : bool, optional
            Flag to apply the noise to the faces of the mesh. The default is False.
        **kwargs :
            Additional keyword arguments specific to the noise model. Common parameters include 'num_octaves' and 'anchor'. Model-specific parameters like 'freq', 'pers', 'slope', 'lacunarity', 'gain', etc., can also be set.

        Returns
        -------
        None
            This method modifies the simulation's surface elevation in-place and does not return a value.

        Notes
        -----
        - The noise is scaled to be volume-conserving, ensuring the mean of the noise is zero.
        - The method internally calculates normalized coordinates based on the target radius and scales the noise appropriately.
        - Default values for noise parameters are set based on the chosen model.
        - For details on thes noise models, see https://www.decarpentier.nl/scape-procedural-basics

        Examples
        --------
        .. code-block:: python
        
            # Apply default turbulence noise:
            sim = cratermaker.Simulation()
            sim.apply_noise()

            #Apply ridged noise with custom Parameters
            sim.apply_noise(model="ridged", noise_width=500e3, num_octaves=10, freq=1.5)
        """      
        scale = self.target.radius / noise_width
        num_octaves = kwargs.pop("num_octaves", 12)
        anchor = kwargs.pop("anchor", self.rng.uniform(0.0,scale, size=(num_octaves, 3))) 
        
        # Set reasonable default values for the different models
        if model == "turbulence" or model == "billowed" or model == "plaw" or model == "ridged":
            kwargs.setdefault("noise_height", noise_height)
            kwargs.setdefault("freq", 2.00)
            kwargs.setdefault("pers", 0.5)
        if model == "plaw":
            kwargs.setdefault("slope", 2.0)
        if model == "swiss" or model == "jordan":
            kwargs.setdefault("lacunarity", 2.00)
            kwargs.setdefault("gain", 0.5)
            kwargs.setdefault("warp", 0.35)
        if model == "jordan":
            kwargs.setdefault("gain0", 0.8)
            kwargs.setdefault("warp0", 0.4)
            kwargs.setdefault("damp0", 1.0)
            kwargs.setdefault("damp", 0.8)
            kwargs.setdefault("damp_scale", 1.0) 
            
        if "noise_height" in kwargs:
            kwargs["noise_height"] = kwargs["noise_height"] / self.target.radius
            
        def _noisemaker(vars):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", Warning)           
                x = self.surf.uxgrid[vars[0]].values * scale 
                y = self.surf.uxgrid[vars[1]].values * scale 
                z = self.surf.uxgrid[vars[2]].values * scale
                noise = apply_noise(model, x, y, z, num_octaves, anchor, **kwargs)
        
            # Make sure the noise is volume-conserving (i.e., the mean is zero)
            # TODO: Take into account the nodes are not uniformly distributed on the sphere
            noise = noise - np.mean(noise)
            return noise                
        
        if to_nodes:
            vars = ['node_x', 'node_y', 'node_z']
            node_noise = _noisemaker(vars)
        else:
            node_noise = None
        if to_faces:
            vars = ['face_x', 'face_y', 'face_z']
            face_noise = _noisemaker(vars)
        else:
            face_noise = None
            
        if model =="swiss" or model == "jordan":
            if node_noise is not None:
                self.surf['node_elevation'] += node_noise * noise_height
            if face_noise is not None:
                self.surf['face_elevation'] += face_noise * noise_height
        else:
            if node_noise is not None:
                self.surf['node_elevation'] += node_noise * self.target.radius 
            if face_noise is not None:
                self.surf['face_elevation'] += face_noise * self.target.radius
        
        return
    
     
    def set_elevation(self, 
                      *args: Any, 
                      **kwargs: Any
                     ) -> None:
        """
        Set the elevation on the surface. Delegates to the Surface object.

        Parameters
        ----------
        *args: Variable length argument list to pass to self.surf.set_elevation.
        **kwargs: Arbitrary keyword arguments to pass to self.surf.set_elevation.
        """
        return self.surf.set_elevation(*args, **kwargs)   


    @property
    def target(self):
        """
        The target body for the impact simulation. Set during initialization.
        """
        return self._target

    @target.setter
    def target(self, value):
        if not isinstance(value, Target):
            raise TypeError("target must be an instance of Target")
        self._target = value

    @property
    def surf(self):
        """
        Surface mesh data for the simulation. Set during initialization.
        """
        return self._surf
    
    @surf.setter
    def surf(self, value):
        if not isinstance(value, Surface):
            raise TypeError("surf must be an instance of Surface")
        self._surf = value

    @property
    def seed(self):
        """
        Seed for the random number generator. Set during initialization.
        """
        return self._seed

    @seed.setter
    def seed(self, value):
        if value is None:
            value = np.random.default_rng().integers(0, 2**32)
        elif not isinstance(value, int):
            raise TypeError("seed must be an integer or None")
        self._seed = value

    @property
    def rng(self):
        """
        Random number generator instance. Set during initialization.
        """
        return self._rng

    @rng.setter
    def rng(self, value):
        if value is None:
            value = np.random.default_rng(seed=self.seed) 
        elif not isinstance(value, Generator): 
            raise TypeError("rng must be a numpy.random.Generator instance or None")
        self._rng = value
    
    @property
    def simdir(self):
        """
        Directory where the simulation data is stored. Set during initialization.
        """
        return self._simdir

    @simdir.setter
    def simdir(self, value):
        if value is None:
            self._simdir = os.getcwd() 
        elif not os.path.isabs(value):
            self._simdir = os.path.abspath(value)
        else:
            self._simdir = value
            
        if not os.path.exists(self._simdir):
            os.makedirs(self._simdir)

    @property
    def production(self):
        """
        The Production class instance used for crater production. Set during initialization.
        """
        return self._production

    @production.setter
    def production(self, value):
        if not issubclass(value.__class__, Production):
            raise TypeError("production must be a subclass of Production")
        self._production = value

    @property
    def scale_cls(self):
        """
        The Scale class that defines the crater scaling law. Set during initialization.
        """
        return self._scale_cls

    @scale_cls.setter
    def scale_cls(self, value):
        if not issubclass(value, Scale):
            raise TypeError("scale_cls must be a subclass of Scale")
        self._scale_cls = value

    @property
    def morphology_cls(self):
        """
        The Morphology class that defines the crater morphology model. Set during initialization.
        """
        return self._morphology_cls

    @morphology_cls.setter
    def morphology_cls(self, value):
        if not issubclass(value, Morphology):
            raise TypeError("morphology_cls must be a subclass of Morphology")
        self._morphology_cls = value

    @property
    def crater(self):
        """
        The current Crater object in the simulation. Set during runtime.
        """
        return self._crater

    @crater.setter
    def crater(self, value):
        if not isinstance(value, Crater):
            raise TypeError("crater must be an instance of Crater")
        self._crater = value

    @property
    def projectile(self):
        """
        The current Projectile object in the simulation. Set during runtime.
        """
        return self._projectile

    @projectile.setter
    def projectile(self, value):
        if value is not None and not isinstance(value, Projectile):
            raise TypeError("projectile must be an instance of Projectile")
        self._projectile = value

    @property
    def data_dir(self):
        """
        Directory where the data files are stored. Dynamically set based on `surf` attribute.
        """
        return self.surf.data_dir

    @property
    def grid_file(self):
        """
        File path of the grid file. Dynamically set based on `surf` attribute.
        """
        return self.surf.grid_file

    @property
    def n_node(self):
        """
        Number of nodes in the simulation mesh. Dynamically set based on `surf` attribute.
        """
        return self.surf.uxgrid.n_node
   
    @property
    def n_node(self):
        """
        Number of nodes in the simulation mesh. Dynamically set based on `surf` attribute.
        """
        return self.surf.uxgrid.n_node

    @property
    def n_face(self):
        """
        Number of faces in the simulation mesh. Dynamically set based on `surf` attribute.
        """
        return self.surf.uxgrid.n_face

    @property
    def interval_number(self):
        """
        The index of the current time step. 
        """
        return self._interval_number

    @interval_number.setter
    def interval_number(self, value):
        if not isinstance(value, int):
            raise TypeError("interval_number must be an integer")
        if value < 0:
            raise ValueError("interval_number must be greater than or equal to zero")
        
        self._interval_number = value    
        
    @property
    def elapsed_time(self):
        """
        The elasped time in My since the start of the simulation.
        """
        return self._elapsed_time
    
    @elapsed_time.setter
    def elapsed_time(self, value):
        self._elapsed_time = np.float64(value)
        
    @property
    def current_age(self):
        """
        The age of the current time step in My relative to the present from the chronology of the production function.
        """
        return self._current_age
    
    @current_age.setter
    def current_age(self, value):
        self._current_age = np.float64(value)
        
    @property
    def elapsed_n1(self):
        """
        The elapsed number of craters larger than 1 km in diameter.
        """
        return self._elapsed_n1
    
    @elapsed_n1.setter
    def elapsed_n1(self, value):
        self._elapsed_n1 = np.float64(value)
       
    @property
    def ejecta_truncation(self):
        """
        The ejecta truncation distance in units of crater radii. Set during initialization.
        """
        return self._ejecta_truncation
    
    @ejecta_truncation.setter
    def ejecta_truncation(self, value):
        if value is not None:
            if not isinstance(value, FloatLike):
                raise TypeError("ejecta_truncation must be a scalar value")
            if value < 1.0:
                raise ValueError("ejecta_truncation must be greater than or equal to zero")
            self._ejecta_truncation = np.float64(value) 
        else:
            self._ejecta_truncation = None
            
    @property
    def smallest_crater(self):
        """
        The smallest crater diameter in meters. Set during initialization.
        """
        return self._smallest_crater
    
    @smallest_crater.setter
    def smallest_crater(self, value):
        if value is None:
            self._smallest_crater = 0.0
            return
        elif not isinstance(value, FloatLike):
            raise TypeError("smallest_crater must be a scalar value")
        elif value < 0:
            raise ValueError("smallest_crater must be greater than or equal to zero")
        elif self._largest_crater is not None and value > self._largest_crater:
            raise ValueError("smallest_crater must be less than or equal to largest_crater")
        self._smallest_crater = np.float64(value)
        
    @property
    def largest_crater(self):
        """
        The largest crater diameter in meters. Set during initialization.
        """
        return self._largest_crater
    
    @largest_crater.setter
    def largest_crater(self, value):
        if value is None or np.isinf(value):
            self._largest_crater = np.inf
            return
        elif not isinstance(value, FloatLike):
            raise TypeError("largest_crater must be a scalar value")
        elif value <= 0:
            raise ValueError("largest_crater must be greater than zero")
        elif self._smallest_crater is not None and value < self._smallest_crater:
            raise ValueError("largest_crater must be greater than or equal to smallest_crater")
        self._largest_crater = np.float64(value)
       
    @property
    def smallest_projectile(self):
        """
        The smallest projectile diameter in meters. Set during initialization.
        """
        return self._smallest_projectile
    
    @smallest_projectile.setter
    def smallest_projectile(self, value):
        if value is None:
            self._smallest_projectile = 0.0
            return
        elif not isinstance(value, FloatLike):
            raise TypeError("smallest_projectile must be a scalar value")
        elif value < 0:
            raise ValueError("smallest_projectile must be greater or equal to zero")
        elif self._largest_projectile is not None and value > self._largest_projectile:
            raise ValueError("smallest_projectile must be less than or equal to largest_projectile")  
        self._smallest_projectile = np.float64(value)
        
    @property
    def largest_projectile(self):
        """
        The largest projectile diameter in meters. Set during initialization.
        """
        return self._largest_projectile
    
    @largest_projectile.setter
    def largest_projectile(self, value):
        if value is None or np.isinf(value):
            self._largest_projectile = np.inf
            return
        elif not isinstance(value, FloatLike):
            raise TypeError("largest_projectile must be a scalar value")
        elif value <= 0:
            raise ValueError("largest_projectile must be greater than zero")
        elif self._smallest_projectile is not None and value < self._smallest_projectile:
            raise ValueError("largest_projectile must be greater than or equal to smallest_projectile")
        self._largest_projectile = np.float64(value)
        
    @property
    def dorays(self):
        """
        Flag to enable the ray pattern model for the ejecta. Set during initialization.
        """
        return self._dorays 
    
    @dorays.setter
    def dorays(self, value):
        if not isinstance(value, bool):
            raise TypeError("dorays must be a boolean value")
        self._dorays = value